program emmsynth_fast
  implicit none

  !EMMSYNTH_FAST uses tensor-product needlets for fast computation of
  ! the values of the magnetic field gradient at the points from the file
  ! 'scattered_points.dat' in a given decimal year and a given magnetic
  ! model (EMM2015 or EMM2017). The gradient is defined in the six files
  ! 'emm2015_*.bin' or 'emm2017_*.bin' (depending on the model), containing
  ! its components weighted values at regular grids on confocal ellipsoids.
  !
  ! INPUT:
  ! 1. FORTRAN unformated data file 'emm201?_*.bin' with the weighted
  !    values at regular grids on confocal ellipsoids.
  ! 2. free format file 'scattered_points.dat' with point's geodetic
  !    coordinates: latitude, longitude (in degrees) and height (in meters).
  ! 3. The chosen magnetic model.
  ! 4. The 'year' when the model is computed.
  !
  ! OUTPUT:
  ! 1. free format file 'scattered_points_values.dat' with point's geodetic
  !    latitude, longitude, (in degrees), height (in meters) and the values
  !    of ten magnetic field components.
  !
  ! Notes.
  !	1. Geodetic coordinates must define points with height
  !    ranging from -415 m to 1,000,000 m.
  ! 2. The 'year' is a decimal number between 2015 and 2020 for EMM2015 or
  !    between 2017 and 2022 for EMM2017.

  ! Version: November 2018
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com> 

  integer :: model, modeldegree
  character(len=80) :: s1, s3
  character(len=80) :: s2_X, s2_XSV, s2_Y, s2_YSV, s2_Z, s2_ZSV
  double precision :: year0, year1, year, dt
  integer*8 icount, icount_rate
  real*8 :: time0, time1, time2, time3, time4, time5
  real*8 :: TTime0, TTime1, TTime2, TTime3, TTime4
  double precision :: ae, inv_flat, be, E, db, b0, pi
  double precision :: NN, tau
  integer, dimension(3) :: ind_zero
  integer :: ndelta3
  integer :: KK, KKext, LLext, LL, LL2
  double precision :: ht, h1, h1i, h2, h2i, h3, h3i
  integer :: NumHeights
  double precision, dimension(10) :: ParKer, ParKer1, ParKer2
  integer, dimension(6) :: ParCub
  integer :: ParMem4
  integer :: ret_i, sizeY1, ndelta1, ndelta2
  double precision, dimension(:), pointer :: Kernel_y1, Kernel_y2
  double precision, dimension(:,:), allocatable :: wb
  double precision, dimension(:,:,:), allocatable :: valueXext
  double precision, dimension(:,:,:), allocatable :: valueYext
  double precision, dimension(:,:,:), allocatable :: valueZext
  double precision, dimension(:,:), allocatable :: pt, values, point
  double precision, dimension(:), allocatable :: phi, phiprime
  double precision :: h_end
  double precision, dimension(:), allocatable :: point3_t
  double precision, dimension(:), allocatable :: point1w, point2w, point3w
  integer :: i, j, k, k1, k2, npt, nu, loopsize, nloop, iloop
  integer, dimension(:), allocatable :: n1, n2, n3
  double precision, dimension(:), allocatable :: t3
  integer, dimension(:,:), allocatable :: index1w, index1t
  integer, dimension(:,:), allocatable :: index2w, index2t
  integer, dimension(:,:), allocatable :: index3w, index3t
  double precision, dimension(:,:), allocatable :: tt1, tt2
  double precision, dimension(:,:), allocatable :: ker_val1, ker_val2, ker_val3
  double precision, dimension(:,:), allocatable :: workX, workY, workZ
  double precision, dimension(:), allocatable :: dphi, cdphi, sdphi
  double precision, dimension(:), allocatable :: Horiz2

  character(len=80) :: debugdir

  parameter (pi = 3.14159265358979323846264338327950288d0)

  nullify(Kernel_y1)

  ! model and year input and check
  write(*, fmt='(A)', advance="no") &
    'Which model to use? (Enter 1 for EMM2015 or 2 for EMM2017): '
  read (*,*) model

  select case(model)
    case(1)
      modeldegree=740
      s2_X='emm2015_Xp.bin'
      s2_XSV='emm2015_XpSV.bin'
      s2_Y='emm2015_Yp.bin'
      s2_YSV='emm2015_YpSV.bin'
      s2_Z='emm2015_Zp.bin'
      s2_ZSV='emm2015_ZpSV.bin'
      year0=2015.0
      year1=2020.0

    case(2)
      modeldegree=790
      s2_X='emm2017_Xp.bin'
      s2_XSV='emm2017_XpSV.bin'
      s2_Y='emm2017_Yp.bin'
      s2_YSV='emm2017_YpSV.bin'
      s2_Z='emm2017_Zp.bin'
      s2_ZSV='emm2017_ZpSV.bin'
      year0=2017.0
      year1=2022.0

    case default
      print *, 'emmsynth_fast:model','wrong input"'
      stop

  end select

  write(*, fmt='(A)', advance="no") &
    'Enter decimal year for geomagnetic computations: '
  read (*,*) year

  if((year.lt.year0).or.(year.gt.year1)) then
    print *, 'emmsynth_fast:year The year must be decimal number between ', &
      year0,' and ', year1
    stop
  end if
  dt=year-year0

  call system_clock(icount, icount_rate)
  time0=dble(icount)/dble(icount_rate)
  TTime1=0
  TTime2=0
  TTime3=0

  ! Earth constants
  ! wgs84_constants
  ae = 6378137d0
  inv_flat=298.257223563d0
  be = ae*(1d0-1d0/inv_flat)
  E=dsqrt((ae-be)*(ae+be))

  ! displacement for the lowest Earth surface point wrt reference ellipsoid
  db=-417d0         !reference ellipsoid displacement in meters $b_0=b_e-417$
  b0=be+db

  ! parameters

  s1='scattered_points.dat' !scattered points input file name
  s3='scattered_points_values.dat' !point and value output file name

  NN=740d0  !grid papameter
  tau=2d0
  ind_zero=(/ 9,9,0 /)    !extended values parameters
  ndelta3=2               !semi-order of Lagrange interpolation

  ! regular grid dimensions
  KK=idnint((1d0+tau/2d0)*NN)+1
  LL2=idnint((1d0+tau/2d0)*NN)
  LL=2*LL2

  h1=pi/(KK-1)
  h1i=1/h1
  h2=pi/LL2
  h2i=1/h2

  ! extended values parameters

  ! extended ellipsoid nodal set dimensions
  KKext=KK+2*ind_zero(1)-1
  LLext=LL+2*ind_zero(2)-1

  ! generating heights
  ht=0.020124185d0
  NumHeights=45

  ! kernel parameters
  ParKer(1)=740d0     !degree of the polynomials reproduced by the needlet kernel
  ParKer(2)=tau       !needlet parameter
  ParKer(3)=0.430729833773221d-5 !target accuracy to be achieved by discrete needlet operator
  ParKer(4)=2d0     ! type of the cutoff function
  ParKer(5)=500d0   ! step refinement parameter
  ParKer(6)=2d0     ! Lagrange interpolation of degree 2*ParKer(6)-1 for fast kernel evaluation
  ParKer(7)=pi      ! radius of the "essential support" of the needlet kernel
  ParKer(8)=0d0     ! "logical", if ParKer(8)=0, then  ParKer(7) is computed in 'kernel_basic'
  ParKer(9)=0d0     ! step of the Lagrangean (almost) equally spaced interpolation knots
  ParKer(10)=1d0    ! number of removed trailing negligible coefficients

  ! cubature parameters
  ParCub(1)=1 ! quadrature type in co-latitude direction
  ParCub(2)=KK ! number of knots in [0,pi] of co-latitude quadrature
  ParCub(3)=LL ! number of equally spaced knots in [0,2*pi) of longitude quadrature
  ParCub(4)=0 ! degree of exactness of the tensor product cubature
  ParCub(5)=0 ! parity of ParCub(2)
  ParCub(6)=2 ! parity of ParCub(3)

  ! memory parameters
  ParMem4=10000  ! number of simultaneously computed polynomial values

  ! formation of parameters for trigonometric needlet kernels
  ParKer1=ParKer !parameters for latitude kernel
  ParKer1(3)=ParKer1(3)/2d0

  sizeY1 = kernel_basic_trig(ParKer1, Kernel_y1) ! allocates Kernel_y1

  ! it is necessary to have ndelta1 <= ind_zero(1)
  ndelta1=floor(ParKer1(7)*h1i)+1

  ! equal kernels in co-latitude and longitude directions
  ParKer2=ParKer1

  allocate(Kernel_y2(sizeY1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: Kernel_y2'
    stop
  end if
  Kernel_y2(1:sizeY1)=Kernel_y1(1:sizeY1)

  ! it is necessary to have ndelta2 <= ind_zero(2)
  ndelta2=floor(ParKer2(7)*h2i)+1

  ! formation of parameters for algebraic Lagrange kernel
  h3=ht
  h3i=1d0/h3

  ! weighted values of the geopotential quantity at regular grid points
  call system_clock(icount)
  time3=dble(icount)/dble(icount_rate)

  allocate(wb(KKext,LLext), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: wb'
    stop
  end if
  
  allocate(valueXext(KKext, LLext, NumHeights), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: valueXext'
    stop
  end if
  valueXext(1:KKext, 1:LLext, 1:NumHeights) = 0d0

  open(10,file=s2_X,form='unformatted',action='read',status='old',iostat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Cannot read data file ', s2_X
    stop
  end if

  do nu=1,NumHeights
    read(10) ((wb(k1,k2), k1=1,KKext), k2=1,LLext)
    valueXext(1:KKext,1:LLext,nu)=wb(1:KKext,1:LLext)
  end do
  close(10)

  open(10,file=s2_XSV,form='unformatted',action='read',status='old', &
    iostat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Cannot read data file', s2_XSV
    stop
  end if

  do nu=1,NumHeights
    read(10) ((wb(k1,k2), k1=1,KKext), k2=1,LLext)
    valueXext(1:KKext,1:LLext,nu)=valueXext(1:KKext,1:LLext,nu) + &
      dt*wb(1:KKext,1:LLext)
  end do
  close(10)

  allocate(valueYext(KKext, LLext, NumHeights), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: valueYext'
    stop
  end if
  valueYext = 0d0

  open(10,file=s2_Y,form='unformatted',action='read',status='old',iostat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Cannot read data file', s2_Y
    stop
  end if

  do nu=1,NumHeights
    read(10) ((wb(k1,k2), k1=1,KKext), k2=1,LLext)
    valueYext(1:KKext,1:LLext,nu)=wb(1:KKext,1:LLext)
  end do
  close(10)

  open(10,file=s2_YSV,form='unformatted',action='read',status='old', &
    iostat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Cannot read data file', s2_YSV
    stop
  end if

  do nu=1,NumHeights
    read(10) ((wb(k1,k2), k1=1,KKext), k2=1,LLext)
    valueYext(1:KKext,1:LLext,nu)=valueYext(1:KKext,1:LLext,nu) + &
       dt*wb(1:KKext,1:LLext)
  end do
  close(10)

  allocate(valueZext(KKext, LLext, NumHeights), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: valueZext'
    stop
  end if
  valueZext = 0d0

  open(10,file=s2_Z,form='unformatted',action='read',status='old',iostat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Cannot read data file', s2_Z
    stop
  end if

  do nu=1,NumHeights
    read(10) ((wb(k1,k2), k1=1,KKext), k2=1,LLext)
    valueZext(1:KKext,1:LLext,nu)=wb(1:KKext,1:LLext)
  end do
  close(10)

  open(10,file=s2_ZSV,form='unformatted',action='read',status='old', &
    iostat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Cannot read data file', s2_ZSV
    stop
  end if

  do nu=1,NumHeights
    read(10) ((wb(k1,k2), k1=1,KKext), k2=1,LLext)
    valueZext(1:KKext,1:LLext,nu)=valueZext(1:KKext,1:LLext,nu) + &
       dt*wb(1:KKext,1:LLext)
  end do
  close(10)

  call system_clock(icount)
  time2=dble(icount)/dble(icount_rate)
  TTime3=TTime3+(time2-time3)

  deallocate(wb)

  ! scattered points for solid spherical polynomial evaluation
  
  npt=dlmsize(s1)

  allocate(pt(npt,3), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: pt'
    stop
  end if

  call dlmread(s1, npt, pt)

  ! checking evaluation point coordinates
  if (any(pt(1:npt,1).lt.-90d0) .or. any(pt(1:npt,1).gt.90d0)) then
    print *, 'emmsynth_fast:pt'
    print *, 'latitude must be between -90 and 90 degrees'
    stop
  end if

  if (any(pt(1:npt,2).lt.-180d0) .or. any(pt(1:npt,2).gt.180d0)) then
    print *, 'emmsynth_fast:pt'
    print *, 'longitude must be between -180 and 180 degrees'
    stop
  end if

  if (any(pt(1:npt,3).lt.-415d0) .or. any(pt(1:npt,3).gt.1d6)) then
    print *, 'emmsynth_fast:pt'
    print *, 'height must be between -415 m and 1,000,000 m'
    stop
  end if

  ! output array
  allocate(values(npt,13), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: values'
    stop
  end if
  values = 0d0
  values(1:npt, 1:3) = pt

  ! work array
  allocate(point(npt,13), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: point'
    stop
  end if

  point=0d0
  point(1:npt,2)=pt(1:npt,2)*pi/180d0
  where (point(1:npt, 2).lt.0d0) point(1:npt,2)=point(1:npt,2)+2d0*pi

  allocate(phi(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: phi'
    stop
  end if
  allocate(phiprime(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: phiprime'
    stop
  end if
  phi=pt(1:npt,1)*pi/180d0
  ![phiprime,point(:,3)]=gd2gc(phi,pt(:,3),ae,be)
  call gd2gc(phi, pt(1:npt,3), ae, be, &
    phiprime, point(1:npt,3))

  point(1:npt,1) = pi/2d0 - phiprime !co-latitude

  deallocate(pt)

  call system_clock(icount)
  time1=dble(icount)/dble(icount_rate)
  TTime2=TTime2+(time1-time2)

  ! evaluation - start

  !from spherical to ellipsoidal coordinates
  call sphere2ellips(point(:,1), point(:,2), point(:,3), E)

  ! checking evaluation point coordinates
  point(:,3)=point(:,3)-b0 !heights
  h_end=floor(b0/(1d0-(ht*dble(NumHeights-ndelta3))**4/4d0)-b0)

  if (any(point(:,3).lt.0) .or. any(point(:,3).gt.h_end)) then
    print *, 'emmsynth_fast:point'
    print *, &
      'the height above reference ellipsoid must be between -0.415 and', &
      (h_end/1d3), 'km'
    stop
  end if

  allocate(point3_t(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: point3_t'
    stop
  end if

  point3_t=(4d0*point(:,3)/(b0+point(:,3)))**(0.25d0)

  allocate(point1w(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: point1w'
    stop
  end if
  allocate(point2w(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: point2w'
    stop
  end if
  allocate(point3w(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: point3w'
    stop
  end if
  allocate(n1(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: n1'
    stop
  end if
  allocate(n2(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: n2'
    stop
  end if
  allocate(n3(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: n3'
    stop
  end if
  allocate(t3(ParMem4), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: t3'
    stop
  end if
  allocate(workX(ParMem4,2*ndelta3), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: workX'
    stop
  end if
  allocate(workY(ParMem4,2*ndelta3), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: workY'
    stop
  end if
  allocate(workZ(ParMem4,2*ndelta3), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: workZ'
    stop
  end if
  allocate(index1w(ParMem4,2*ndelta1), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: index1w'
    stop
  end if
  allocate(index1t(ParMem4,2*ndelta1), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: index1t'
    stop
  end if

  allocate(tt1(ParMem4,2*ndelta1), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: tt1'
    stop
  end if
  allocate(ker_val1(ParMem4,2*ndelta1), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: ker_val1'
    stop
  end if
  allocate(index2w(ParMem4,2*ndelta2), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: index2w'
    stop
  end if
  allocate(index2t(ParMem4,2*ndelta2), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: index2t'
    stop
  end if
  allocate(tt2(ParMem4,2*ndelta2), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: tt2'
    stop
  end if
  allocate(ker_val2(ParMem4,2*ndelta2), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: ker_val2'
    stop
  end if
  allocate(index3w(ParMem4,2*ndelta3), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: index3w'
    stop
  end if
  allocate(index3t(ParMem4,2*ndelta3), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: index3t'
    stop
  end if
  allocate(ker_val3(ParMem4,2*ndelta3), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: ker_val3'
    stop
  end if

  ! evaluation - general loop
  nloop=ceiling(dble(npt)/dble(ParMem4))
  loopsize = ParMem4

  do iloop=1,nloop
    if (iloop.lt.nloop) then
      point1w(1:loopsize)=point((iloop-1)*ParMem4+1:iloop*ParMem4,1)
      point2w(1:loopsize)=point((iloop-1)*ParMem4+1:iloop*ParMem4,2)
      point3w(1:loopsize)=point3_t((iloop-1)*ParMem4+1:iloop*ParMem4)
    else
      ! on the last pass, adjust loopsize to handle the smaller chunk
      ! of the batch that remains
      loopsize = npt - (iloop-1)*ParMem4
      point1w(1:loopsize)=point((iloop-1)*ParMem4+1:npt,1)
      point2w(1:loopsize)=point((iloop-1)*ParMem4+1:npt,2)
      point3w(1:loopsize)=point3_t((iloop-1)*ParMem4+1:npt)
    end if

  ! neighbourhood and distances formation
    select case(ParCub(1))
      case(1)
        n1(1:loopsize)=floor(point1w(1:loopsize)*h1i)
        do i = 1, loopsize
          do j = 1, 2*ndelta1
            index1w(i,j) = j + n1(i) - ndelta1
            index1t(i,j) = index1w(i,j) + ind_zero(1)
            tt1(i,j) = dabs(index1w(i,j)*h1 - point1w(i))
          end do
        end do

      case(2)
        n1(1:loopsize) = floor(point1w(1:loopsize)*h1i-0.5d0)
        do i = 1, loopsize
          do j = 1, 2*ndelta1
            index1w(i,j) = j + n1(i) - ndelta1
            index1t(i,j) = index1w(i,j) + ind_zero(1) + 1
            tt1(i,j) = dabs((index1w(i,j)+0.5d0)*h1 - point1w(i))
          end do
        end do
    end select

    n2(1:loopsize)=floor(point2w(1:loopsize)*h2i)
    do i = 1, loopsize
      do j = 1, 2*ndelta2
        index2w(i,j) = j + n2(i) - ndelta2
        index2t(i,j) = index2w(i,j) + ind_zero(2)
        tt2(i,j) = dabs(index2w(i,j)*h2 - point2w(i))
      end do
    end do

    n3(1:loopsize) = floor(point3w(1:loopsize)*h3i)
    t3(1:loopsize) = point3w(1:loopsize)*h3i-n3(1:loopsize)
    do i = 1, loopsize
      do j = 1, 2*ndelta3
        index3w(i,j) = j + n3(i) - ndelta3
        index3t(i,j) = abs(index3w(i,j)+ind_zero(3))+1
      end do
    end do


    ! kernel evaluations
    call kernel_values(ParKer1, Kernel_y1, (2*ndelta1*loopsize), &
      tt1(1:loopsize,1:2*ndelta1), ker_val1(1:loopsize,1:2*ndelta1))
    call kernel_values(ParKer2, Kernel_y2, (2*ndelta2*loopsize), &
      tt2(1:loopsize,1:2*ndelta2), ker_val2(1:loopsize,1:2*ndelta2))
    ker_val3(1:loopsize,1:2*ndelta3) &
      = lagrange_basis(2*ndelta3, t3(1:loopsize))

  ! needlet evaluation
    workX(1:loopsize, 1:2*ndelta3) = 0d0
    workY(1:loopsize, 1:2*ndelta3) = 0d0
    workZ(1:loopsize, 1:2*ndelta3) = 0d0
    do k=1,loopsize
     do j=1,2*ndelta3
      workX(k,j) = &
       dot_product( &
        matmul( &
         ker_val1(k,1:2*ndelta1), &
         valueXext(index1t(k,1:2*ndelta1),index2t(k,1:2*ndelta2),index3t(k,j))&
        ), &
        ker_val2(k,1:2*ndelta2) &
       )
      workY(k,j) = &
       dot_product( &
        matmul( &
         ker_val1(k,1:2*ndelta1), &
         valueYext(index1t(k,1:2*ndelta1),index2t(k,1:2*ndelta2),index3t(k,j))&
        ), &
        ker_val2(k,1:2*ndelta2) &
       )
      workZ(k,j) = &
       dot_product( &
        matmul( &
         ker_val1(k,1:2*ndelta1), &
         valueZext(index1t(k,1:2*ndelta1),index2t(k,1:2*ndelta2),index3t(k,j))&
        ), &
       ker_val2(k,1:2*ndelta2) &
      )
     end do
    end do
 
    values((iloop-1)*ParMem4+1:(iloop-1)*ParMem4+loopsize, 4) = &
        sum(workX(1:loopsize,1:2*ndelta3)*ker_val3(1:loopsize,1:2*ndelta3),2)
    values((iloop-1)*ParMem4+1:(iloop-1)*ParMem4+loopsize, 5) = &
        sum(workY(1:loopsize,1:2*ndelta3)*ker_val3(1:loopsize,1:2*ndelta3),2)
    values((iloop-1)*ParMem4+1:(iloop-1)*ParMem4+loopsize, 6) = &
        sum(workZ(1:loopsize,1:2*ndelta3)*ker_val3(1:loopsize,1:2*ndelta3),2)
  end do !iloop

  deallocate(ker_val3)
  deallocate(index3t)
  deallocate(index3w)
  deallocate(ker_val2)
  deallocate(tt2)
  deallocate(index2t)
  deallocate(index2w)
  deallocate(ker_val1)
  deallocate(tt1)
  deallocate(index1t)
  deallocate(index1w)
  deallocate(workZ)
  deallocate(workY)
  deallocate(workX)
  deallocate(t3)
  deallocate(n3)
  deallocate(n2)
  deallocate(n1)
  deallocate(point3w)
  deallocate(point2w)
  deallocate(point1w)
  deallocate(point3_t)
  deallocate(point)
  deallocate(valueZext)
  deallocate(valueYext)
  deallocate(valueXext)
  deallocate(Kernel_y2)

  ! computing quantities in the ellipsoidal reference frame

  allocate(dphi(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: dphi'
    stop
  end if

  allocate(cdphi(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cdphi'
    stop
  end if

  allocate(sdphi(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: sdphi'
    stop
  end if

  ! rotating geocentric to ellipsoidal reference frame
  dphi=(phiprime-phi)
  cdphi=dcos(dphi)
  sdphi=dsin(dphi)
  values(1:npt,7)=values(1:npt,4)*cdphi-values(1:npt,6)*sdphi !X north component
  values(1:npt,8)=values(1:npt,5)                             !Y east component
  values(1:npt,9)=values(1:npt,4)*sdphi+values(1:npt,6)*cdphi !Z down component

  deallocate(sdphi)
  deallocate(cdphi)
  deallocate(dphi)

  deallocate(phiprime)
  deallocate(phi)

  ! evaluating the remaining components of the field
  allocate(Horiz2(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: Horiz2'
    stop
  end if

  Horiz2=values(:,7)**2+values(:,8)**2
  values(:,10)=dsqrt(Horiz2)                   ! H horizontal intensity
  values(:,11)=dsqrt(Horiz2+values(:,9)**2)    ! F total intensity
  values(:,12)=datan2(values(:,9),values(:,10))! I inclination
  values(:,13)=datan2(values(:,8),values(:,7)) ! D declination

  deallocate(Horiz2)

  call system_clock(icount)
  time4=dble(icount)/dble(icount_rate)
  TTime1=TTime1+(time4-time1)

  ! evaluation - end

  ! output

  open(unit=1, file=s3, status='replace', iostat=ret_i)
  if(ret_i .ne. 0) then
    print *, 'Cannot write output file ', s3
    stop
  end if

  do k = 1,npt
    write(1, '(2(F20.14),F16.8,8(F12.4),2(F12.8))') ( values(k,i), i=1,13 )
  end do
      
  close(1)

  call system_clock(icount)
  time5=dble(icount)/dble(icount_rate)

  TTime4=time5-time4
  TTime0=time5-time0

  ! statistics

  write (*,'(A)') &
    '---------------------------------------------------------------'
  select case(model)
    case(1)
      write(*,'(A)') '        Program:  emmsynth_fast,       Model: EMM2015'
    case(2)
      write(*,'(A)') '        Program:  emmsynth_fast,       Model: EMM2017'
    case default
      write(*,'(A)') '        Program:  emmsynth_fast,       Model: UNKNOWN'
  end select
  write (*,'(A,F10.5)') &
    '   Evaluation year:            = ', year
  write (*,'(A,A)') &
    '---------------------------------------------------------------'
  write (*,'(A,F15.5,A)') &
    '   Total time                  = ', TTime0, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    '   Grid values load time       = ', TTime3, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    '   Coordinates load time       = ', TTime2, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    '   Synthesis time              = ', TTime1, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    '   Values write time           = ', TTime4, ' CPU seconds'
  write (*,'(A,I9)') &
    '   Total number of points      = ', npt
  write (*,'(A,F15.5)') &
    '   Retrieved points per second = ', (real(npt)/TTime1)
  write (*,'(A)') &
    '---------------------------------------------------------------'

  deallocate(values)

  contains

  !KERNEL_BASIC_TRIG computes the values of the trigonometric kernel at
  !equally spaced knots
  !
  ! The code evaluate via Clenshaw recurrence the even trigonometric kernel
  ! $y(x)=\sum_{n=0}^\infty \varphi(n/N) \cos(nx)$
  ! for $\varphi(0)=.5$; $\varphi(t)=1, 0<t\le 1$; and $0\le\varphi(t)\le 1$,
  ! $1<t\le 1+\tau$, where $N=ParKer(1)$ and $\tau=ParKer(2)$.
  ! The type of the cutoff function $\varphi$ is given by ParKer(4).
  ! The knots  X, X(k)=(k-nu)*h, k=1,2,..., (nu=ParKer(6))   are equally
  ! spaced with step  h=ParKer(9) in the interval [(1-nu)*h,delta+nu*h]
  ! and delta=ParKer(7).
  !
  ! INPUT:
  ! 1. ParKer -- see Parmeters. The code uses all elements of  ParKer.
  ! OUTPUT:
  ! 1. ParKer -- see Parmeters
  ! 2. Y - vector with the values of the kernel at the knots  X
  ! 3. X - vector with the values of the knots
  !
  ! Notes.
  ! 1. Y  is of the same length as  X.
  ! 2. ParKer is an output variable because the code defines
  !   a) ParKer(9) as ParKer(9)=pi/(Degree*ParKer(5) )with
  !      Degree=ParKer(1)+ceil(ParKer(1)*ParKer(2))-ParKer(10)
  !   b) ParKer(7) (if ParKer(8)=0).
  ! 3. If ParKer(8)=0 then ParKer(7) is the output of 'length_supp_trig' plus
  ! pi/(Degree+ParKer(1)).
  !
  ! Version: October 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>
  function kernel_basic_trig(ParKer, Y) result(ny)

    integer :: ny, k0, N, Degree, k, Np0, i, ret_i
    double precision, dimension(:), intent(inout) :: ParKer
    double precision, dimension(:), pointer, intent(out) :: Y
    double precision, dimension(:), allocatable :: X, wc, Coef, Coef0
    double precision :: pi, m, b

    parameter (pi = 3.14159265358979323846264338327950288d0)

    k0 = int(ParKer(1))
    N = ceiling(ParKer(1)*ParKer(2))
    Degree = N+k0-int(ParKer(10))  ! total degree of polynomial kernel
    m = -dlog10(ParKer(3))

    allocate(wc(N+1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: wc'
      stop
    end if

    allocate(Coef(Degree+1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: Coef'
      stop
    end if

    ! coefficients formation
    select case (int(ParKer(4)))
      case (1)
        b = dble(floor(0.8722d0 * m - 1.5416d0 + 0.5d0))
        call cut_off(N,1,b,wc)

      case (2)
        b = (58d0*m-7d0)/25d0
        call cut_off(N,2,b,wc)

    end select

    Coef = 1d0
    Coef(max(k0,0)+1:Degree+1) = wc(N+min(k0,0)+1:int(ParKer(10))+1:-1)
    Coef(1) = Coef(1)/2d0

    ! evaluation of delta
    if (ParKer(8) .eq. 0d0) then
      ParKer(7) = length_supp_trig(Coef,ParKer(3),3) + pi/dble(k0+Degree)
    else
      ParKer(7) = length_supp_trig(Coef,ParKer(3),3) 
    end if

    deallocate(wc)

    ! computation of basic values
    ParKer(9) = pi/dble(Degree*ParKer(5))  ! the grid step;
    Np0 = ceiling(ParKer(7)/ParKer(9))

    ny = Np0 + 2*int(ParKer(6))
    allocate(X(ny), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: X'
      stop
    end if

    do i = 1,ny
      X(i) = dble(i-ParKer(6)) * ParKer(9) ! equally spaced knots
    end do


    allocate(Coef0(Degree+1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: Coef0'
      stop
    end if

    Coef0(1:Degree+1) = 0d0

    call clenshaw_trig(Coef, Coef0, X, Y)

    do k = 1, int(ParKer(6))-1
      Y(k) = Y(2*int(ParKer(6))-k)
    end do

    deallocate(X)
    deallocate(Coef0)
    deallocate(Coef)

  end function kernel_basic_trig

  !CUT_OFF computes the values of a cutoff function
  !
  ! The continuous cutoff function  $y(t)=\phi(t)$ satisfies
  ! $\phi : [0,1]\to[0,1]$, $\phi(0)=0, \phi(1)=1$.
  !
  ! INPUT:
  ! 1. R is the number of equal intervals on which [0,1] is divided
  !    in order to form the knots t(k)=(k-1)/R, k=1,2,...,R+1.
  ! 2. tip  is the type of the generating cutoff function,
  ! 3. b  is a parameter of the cutoff function.
  ! tip = 1: $\phi(t)=\int_0^t (\pi/2)*\sin^{2*b+1}(\pi v) dv *(2*b+1)!!/(2*b)!!$
  ! tip = 2: $\phi(t)=\kappa \int_0^t \exp{2*b*\sqrt{v*(1-v)}} dv$
  !          with $\kappa =( \int_0^1 \exp{2*b*\sqrt{v*(1-v)}} dv)^{-1}$.
  !
  ! OUTPUT:
  ! 1. y  the values of the cutoff function at the knots  t.
  !
  ! Notes.
  ! 1. In the spherical case for:
  !  a) tip=1 set  b(m)=4,5,5,6,7,8 for m=5,6,7,8,9,10, eps=10^(-m).
  !  b) tip=2 set $b(m)=2.4*m+1.7-0.1*\tau+0.1*(\tau-3)_+$ for m=5,6,7,8,9,10, eps=10^(-m)
  ! 2. In the trigonometric case for
  !  a) tip=1 set
  !  b) tip=2 set $b(m)=(58*m-7)/25$ for m=5,6,7,8,9,10,11, eps=10^(-m)

  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  subroutine cut_off(R,tip,b, y)

  integer, intent(in) :: R, tip
  double precision, intent(in) :: b
  double precision, intent(out) :: y(R+1)

  double precision :: S2(R+1), y0(R+1), y2(int(R/2)+1)
  double precision :: v, w, x, pi
  integer :: i, j, k, R2

  parameter (pi = 3.14159265358979323846264338327950288d0)

  v =  dble(R)
  R2 = int(R/2)

  select case (tip)
    case (1)
      do i=1,R
        S2(i) = (dsin(pi*dble(i-1)/v))**2
      end do
      do i=1,R+1
        y(i) = 1d0
      end do
      do k = int(b),1,-1
        do i=1,R+1
          y(i) = 1d0 + y(i)*S2(i)*(1d0-1d0/dble(2*k))
        end do
      end do
    
      do i=1,R
        y(i) = 0.5d0 - y(i)*dcos(pi*dble(i-1)/v)/2d0;
        if(y(i) .lt. 0d0) y(i) = 0d0
        if(y(i) .gt. 1d0) y(i) = 1d0
      end do

      x = 1d0
      y(R+1) = 1d0
      do k = R,1,-1
        if (y(k) .gt. x) y(k)=x
        x = y(k)
      end do

      y(1) = 0d0

      if (2*R2 .eq. R) then
        y(R2+1) = 0.5d0
        y(R+1:R2+2:-1) = 1d0-y(1:R2)
        y(1:R2) = 1d0-y(R+1:R2+2:-1)
      else
        y(R+1:R2+2:-1) = 1d0-y(1:R2+1)
        y(1:R2+1) = 1d0-y(R+1:R2+2:-1)
      end if
    case (2)
      do i = 1,R+1
        y0(i) = dexp(b*dsqrt(1d0 - (dble(i-1)/v)**2))
      end do

      if(R .ne. 2*R2) then
        y2(1) = 2d0*y0(1) + y0(2)
      else
        y2(1) = 0d0
      end if

      j = R+1-2*R2
      do i=2,R2+1
        y2(i) = (y0(j) + 4d0*y0(j+1) + y0(j+2)) + y2(i-1)
        j = j+2
      end do

      w = 2d0 * y2(R2+1)
      do i = 1,R2+1
        v = y2(i) / w
        y(R-R2+i) = 0.5d0 + v
        y(R2+2-i) = 0.5d0 - v
      end do
    end select
  end subroutine cut_off

  !LENGTH_SUPP_TRIG finds the interval [0,delta], where an even polynomial is
  !supported. 
  !
  ! For $f(t)=\sum_{k=0}^Degree C_k \cos(kt)$, Degree=length(C),
  ! the function gives the smallest number  delta, such that
  ! $\int_\delta^\pi |f| \le Epsilon |\int_0^\pi f| = \pi*|C_0|*Epsilon$.
  !
  ! INPUT: 
  ! 1. C is the polynomial coefficient vector.
  ! 2. Epsilon is the accuracy of the "support".
  ! 3. Digits is the the number of significant digits in  delta.
  !
  ! OUTPUT: 
  ! 1.delta  is the the length of the "support".
  !
  ! Notes. 
  ! 1. If C(1)=0, then  delta  is the smallest number satisfying 
  ! $\int_\delta^{\pi} |f(t)|\,dt\le \pi\epsilon$.
  ! 2. The following conditions has to be met: 
  !  a) 1 <= Degree <= 1000000
  !  b) Epsilon is real, 0 < Epsilon < 1
  !  c) Digits is an integer, 2 <= Digits <= 5. Usually Digits=3.

  ! Version: October 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>
  
  function length_supp_trig(C,Epsilon,Digits) result(delta)

    double precision, dimension(:), intent(in) :: C
    double precision, intent(in) :: Epsilon
    integer, intent(in) :: Digits
    double precision :: delta

    double precision, dimension(size(C)) :: S
    double precision, dimension(:), allocatable :: x
    double precision, dimension(:), pointer :: y
    integer :: Degree, m, kk, k, i, ret_i
    double precision :: thresh, pi, kappa, h0, logh0, a0, sum0, loga0
    logical :: cond

    parameter (pi = 3.14159265358979323846264338327950288d0)

    nullify(y)

    S(:) = 0d0

    Degree = size(C)
    if((Degree .lt. 1) .or. (Degree .gt. 1000000)) then
      print *, 'LENGTH_SUPP_TRIG:Degree'
      print *, 'The polynomial degree must be less than 1,000,000.'
      stop
    end if
    if((Epsilon .le. 0d0) .or. (Epsilon .ge. 1d0)) then
      print *, 'LENGTH_SUPP_TRIG:Epsilon'
      print *, 'Presision Epsilon is a real number between 0 and 1.'
    end if
    if((Digits .lt. 2) .or. (Digits .gt. 5)) then
      print *, 'LENGTH_SUPP_TRIG:Digits'
      print *, 'The number of significant digits is between 2 and 5.'
    end if

    thresh = pi*Epsilon

    if (C(1) .ne. 0) thresh = thresh*dabs(C(1))
    kappa = 2d0*dlog(dble(Degree)/Epsilon)/dlog(2d0)

    h0 = 10d0**dble(-Digits-1)
    logh0 = dlog(1d0-h0)
    m = int(dlog(0.5d0)/logh0)

    allocate(x(m+1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: x'
      stop
    end if

    a0 = pi
    cond = .true.
    sum0 = 0d0

    kk = 0
    do while(cond .and. (kk .lt. kappa))
      kk = kk+1
      loga0 = dlog(a0);
      do i = 1,m+1
        x(i) = dexp(dble(i-1)*logh0+loga0)
      end do 
      call clenshaw_trig(C,S,x, y)
      do i = 1,m+1
        y(i) = dabs(y(i))
      end do
    
      sum0 = sum0 + y(1)*(x(1)-x(2))/2d0 + y(m+1)*(x(m)-x(m+1))/2d0
      sum0 = sum0 + sum(y(2:m)*(x(1:m-1)-x(3:m+1)))/2d0
      if (sum0 .gt. thresh) then
        cond = .false.
      else
        a0 = x(m+1)
        deallocate(y)
      end if
    end do

    k = m+2
    do while (sum0 .gt. thresh)
      k = k-1
      sum0 = sum0 -(y(k-1)+y(k))*(x(k-1)-x(k))/2d0
    end do
    deallocate(y)

    if (kk .lt. kappa) then
      delta = x(k-1)
    else
      delta = 0d0
    end if

    deallocate(x)

  end function length_supp_trig

  !CLENSHAW_TRIG evaluates trigonometric polynomials by their coefficients
  ! $y(\theta)=\sum_{r=0}^{N-1} (C_r\cos(r\theta)+S_r\sin(r\theta))$.
  ! using Newbery modification of Clenshaw recurrence [1].
  ! INPUT: C, S - polynomial coefficient vectors
  !        theta - matrix of the independent variable
  ! OUTPUT: y - matrix with the polynomial values
  ! 
  ! Notes: 1. C  and  S  are vectors of the same length. 
  !           C(r+1)  contains $C_r$, S(r+1) contains $S_r$. 
  ! 2. Polynomial degree  N-1  is calculated as \verb#length(C)-1#.
  ! 3. S(1)  is not used in computation because $S_0\sin(0\theta)=0$
  ! 4. y  is of the same size as  theta.
  !
  ! [1] A. C. R. Newbery, Error Analysis for Fourier Series Evaluation,
  ! Mathematics of Computation, 27, 123, (1973), 639-644

  ! Version: October 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  subroutine clenshaw_trig(C,S,theta, y)

    double precision, dimension(:), intent(in) :: C, S, theta
    double precision, dimension(:), pointer, intent(out) :: y

    double precision, dimension(size(C)) :: C2, S2

    integer :: N, i, j, sz, m, ret_i
    double precision, dimension(:), allocatable :: t1, t2, y1, y2, ct1, ct2
    double precision, dimension(:), pointer :: t0, u1, u2, u3

    logical, dimension(:), allocatable :: ind1

    double precision :: pi
    parameter (pi = 3.14159265358979323846264338327950288d0)

    ! coefficients check
    N = size(C)
    if (N .ne. size(S)) then
      print *,'CLENSHAW_TRIG:coefficients'
      print *,'Coefficient vectors C and S must be of the same length!'
    end if

    ! variable change 0

    ! This allocation (t0) will also be returned as Y. The calling routine
    ! is responsible for deallocation
    nullify(t0)

    if(associated(y)) then
      if(size(y) .eq. size(theta)) then
        t0 => y
      else
        deallocate(y)
      end if
    end if

    if(.not. associated(t0)) then
      allocate(t0(size(theta)), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: t0'
        stop
      end if
    end if

    do i = 1, size(theta)
      t0(i) = dmod(theta(i), 2d0*pi)
    end do

    ! variable change 1
    allocate(ind1(size(theta)), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: ind1'
      stop
    end if

    sz = 0
    do i=1,size(theta)
      if (dabs(dsin(t0(i))) .ge. dsin(pi/4d0)) then
        ind1(i) = .true.
        sz = sz + 1
      else
        ind1(i) = .false.
      end if
    end do

    allocate(t1(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: t1'
      stop
    end if

    allocate(u1(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: u1'
      stop
    end if

    allocate(u2(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: u2'
      stop
    end if

    allocate(ct1(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: ct1'
      stop
    end if

    j = 1
    do i=1,size(theta)
      if(ind1(i)) then
        t1(j) = t0(i)
        j = j + 1
      end if
    end do

    ! Clenshaw reccurence 1
    do i=1,sz
      u1(i) = 0d0
      u2(i) = 0d0
      ct1(i) = dcos(t1(i))
    end do

    do m = N,2,-1
      u3 => u2
      u2 => u1
      u1 => u3
      !u1(:) = (C(m) + 2d0*ct1(:)*u2(:)) - u3(:)
      do i=1,sz
        u1(i) = (C(m) + 2d0*ct1(i)*u2(i)) - u3(i)
      end do
    end do

    allocate(y1(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: y1'
      stop
    end if

    do i=1,sz
      y1(i) = (C(1) + ct1(i)*u1(i)) - u2(i)
    end do

    do i=1,sz
      u1(i) = 0d0
      u2(i) = 0d0
    end do

    do m = N,2,-1
      u3 => u2
      u2 => u1
      u1 => u3
      !u1(:) = (S(m) + 2d0*ct1(:)*u2(:)) - u3(:)
      do i=1,sz
        u1(i) = (S(m) + 2d0*ct1(i)*u2(i)) - u3(i)
      end do
    end do

    deallocate(ct1)

    y => t0

    j = 1
    do i=1, size(theta)
      if(ind1(i)) then
        y(i) = y1(j) + dsin(t1(j))*u1(j)
        j = j+1
      end if
    end do

    deallocate(y1)
    deallocate(u2)
    deallocate(u1)
    deallocate(t1)

    ! variable change 2
    !ind2=~ind1;
    sz = size(theta) - sz

    allocate(t2(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: t2'
      stop
    end if

    !t2 = pack(t0,.not. ind1)
    j = 1
    do i = 1,size(theta)
      if(.not. ind1(i)) then
        t2(j) = t0(i)
        j = j + 1
      end if
    end do

    C2(:)      =  C(:)
    S2(:)      = -S(:)

    C2(2:size(C):4) =  S(2:size(C):4)
    S2(2:size(C):4) =  C(2:size(C):4)

    C2(3:size(C):4) = -C(3:size(C):4);
    S2(3:size(C):4) =  S(3:size(C):4);

    C2(4:size(C):4) = -S(4:size(C):4);
    S2(4:size(C):4) = -C(4:size(C):4);

    ! Clenshaw reccurence 2

    allocate(u1(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: u1'
      stop
    end if

    allocate(u2(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: u2'
      stop
    end if

    allocate(ct2(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: ct2'
      stop
    end if

    allocate(y2(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: y2'
      stop
    end if

    do i=1,sz
      u1(i) = 0d0
      u2(i) = 0d0

      ct2(i) = dsin(t2(i))
    end do

    do m = N,2,-1
      u3 => u2
      u2 => u1
      u1 => u3

      !u1(:) = (C2(m) + 2d0*ct2(:)*u2(:)) - u3(:)
      do i=1,sz
        u1(i) = (C2(m) + 2d0*ct2(i)*u2(i)) - u3(i)
      end do
    end do

    do i=1,sz
      y2(i) = (C2(1) + ct2(i)*u1(i)) - u2(i)
    end do

    do i=1,sz
      u1(i) = 0d0
      u2(i) = 0d0
    end do

    do m = N,2,-1
      u3 => u2
      u2 => u1
      u1 => u3
      do i=1,sz
        u1(i) = (S2(m) + 2d0*ct2(i)*u2(i)) - u3(i)
      end do
    end do

    j = 1
    do i=1, size(theta)
      if(.not. ind1(i)) then
        y(i) = y2(j) + dcos(t2(j))*u1(j)
        j = j+1
      end if
    end do

    deallocate(ct2)
    deallocate(u2)
    deallocate(u1)
    deallocate(t2)
    deallocate(ind1)

  end subroutine clenshaw_trig

  !SPHERE2ELLIPS converts the spherical coordinates (theta,lambda,r)
  ! of a 3-d point to ellipsoidal coordinates (phi,lambda,u).
  ! 'e' is the linear eccentricity of the confocal ellipses.
  ! e^2=a^2-b^2, where 'a', 'b' are the semi-axes.
  ! Forward stability for points with  u>e.
  ! Requirements: The arrays 'theta', 'lambda', 'r' must be of the same size
  ! '0<=theta<=pi', '0<=lambda<2 pi', 'r>=0' 'e' must be non-negative.
  !
  ! Returns in place: theta <= phi, lambda <= lambda, r <= u
  !
  subroutine sphere2ellips(theta,lambda,r,e)
    double precision, dimension(:), intent(inout) :: theta, lambda, r
    double precision, intent(in) :: e

    double precision, dimension(size(r)) :: s

    double precision :: pi
    parameter (pi = 3.14159265358979323846264338327950288d0)

    if ((size(theta).ne.size(lambda)).or.(size(theta).ne.size(r))) then
      print *, 'sphere2ellips:input'
      print *, '"theta", "lambda" and "r" must be of the same size'
      stop
    end if

    if(.not.(e.ge.0)) then
      print *, 'sphere2ellips:e'
      print *, '"e" must be non-negative'
    end if

    if(any(theta.gt.pi).or.any(theta.lt.0d0)) then
      print *, 'sphere2ellips:theta'
      print *, '"theta" must be between 0 and pi'
      stop
    end if

    if(any(lambda.ge.2d0*pi).or.any(lambda.lt.0d0)) then
      print *, 'sphere2ellips:lambda'
      print *, '"lambda" must be between 0 and 2*pi'
      stop 
    end if

    if(any(r.lt.0)) then
      print *,'sphere2ellips:r'
      print *,'"r" must be non-negative'
      stop
    end if

    s(:)=(r(:)**2-e**2)/2d0
    r(:)=dsqrt(s(:)+dsqrt(s(:)**2+e**2*r(:)**2*dcos(theta(:))**2))
    theta(:)=datan2(dsin(theta(:)),dsqrt(1d0+e**2/r(:)**2)*dcos(theta(:)))

  end subroutine sphere2ellips

  !KERNEL_VALUES computes the values of a needlet kernel considered as
  ! an even smooth function "supported" in a neighbourhood of the origin.
  !
  ! The code uses 2*ParKer(6)-1 degree Lagrange interpolation at 
  ! equally spaced knots with step in ParKer(9)
  ! as needlet kernel values are given in 'Kernel_y'.
  !
  ! INPUT:
  ! 1. ParKer -- see Parameters
  ! 2. Kernel_y  is the vector with kernel values at equally spaced nodes,
  !    Kernel_y(k) contains the value at the node  (k-ParKer(6))*ParKer(9).
  ! 3. x  is an array with the values of the independent variable.
  !
  ! OUTPUT: 
  ! 1. y  is an array with the values of the kernel
  !    y is of the same size as  x.
  !
  ! Notes. 
  ! 1. The code uses  ParKer(6), ParKer(7), ParKer(9). The dependence of
  ! the kernel on the other elements of  ParKer  is through 
  ! the values of  Kernel_y.
  ! 2. For x>ParKer(7) the code returns y=0.
  ! 3. The code assumes but does not check that x>=0.

  ! Version: October 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  subroutine kernel_values(ParKer,Kernel_y, sizey, x, y)
    double precision, dimension(:), intent(in) :: ParKer, Kernel_y
    integer, intent(in) :: sizey
    double precision, dimension(sizey), intent(in) :: x
    double precision, dimension(sizey), intent(out) :: y

    integer :: sz
    double precision, dimension(:), allocatable :: x0
    logical, dimension(:), allocatable :: ind
    integer, dimension(:), allocatable :: k
    double precision, dimension(:), allocatable :: d, yy, c
    double precision, dimension(:,:), allocatable :: zz

    integer :: nu, nu2, length, m, mu, n, ret_i, i, j

    y(1:sizey) = 0d0

    allocate(ind(sizey), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: ind'
      stop
    end if

    ind(1:sizey) = (x(1:sizey) .le. ParKer(7))
    sz = count(x(1:sizey) .le. ParKer(7))

    allocate(x0(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: x0'
      stop
    end if

    allocate(k(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: k'
      stop
    end if

    allocate(d(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: d'
      stop
    end if

    allocate(yy(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: yy'
      stop
    end if

    j=1
    do i=1,sizey
      if(ind(i)) then
        x0(j) = x(i)
        j = j + 1
      end if
    end do

    k(1:sz) = int(x0(1:sz)/ParKer(9)) + 1
    d(1:sz) = x0(1:sz)/ParKer(9) + 1 - k

    select case(int(ParKer(6)))
      case(2)
        yy(1:sz) = &
          (-d(1:sz)*Kernel_y(k) + &
            (d(1:sz)+1)*Kernel_y(k+1)*3d0)*(d(1:sz)-1)*(d(1:sz)-2) + &
          (-(d(1:sz)-2)*Kernel_y(k+2)*3d0 + &
            (d(1:sz)-1)*Kernel_y(k+3))*(d(1:sz)+1)*d(1:sz)
        yy(1:sz) = yy(1:sz)/6d0

      case(3)
        yy(1:sz) = &
          ( &
            -(d(1:sz)+1)*d(1:sz)*Kernel_y(k) &
            +(d(1:sz)+2)*d(1:sz)*Kernel_y(k+1)*5d0 &
            -(d(1:sz)+2)*(d(1:sz)+1)*Kernel_y(k+2)*10d0 &
          )*(d(1:sz)-1)*(d(1:sz)-2)*(d(1:sz)-3) + &
          ( &
             (d(1:sz)-2)*(d(1:sz)-3)*Kernel_y(k+3)*10d0 &
            -(d(1:sz)-1)*(d(1:sz)-3)*Kernel_y(k+4)*5d0 &
            +(d(1:sz)-1)*(d(1:sz)-2)*Kernel_y(k+5) &
          )*(d(1:sz)+2)*(d(1:sz)+1)*d(1:sz)
        yy(1:sz) = yy(1:sz)/120d0

      case(1)
        yy(1:sz) = (1-d(1:sz))*Kernel_y(k) + d(1:sz)*Kernel_y(k+1)

      case default
        nu = int(ParKer(6))
        nu2 = nu*2
        length = sz

        allocate(zz(length,nu2), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: zz'
          stop
        end if

        allocate(c(length), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: c'
          stop
        end if

        zz(1:length,1:nu2) = 0d0

        do n = 1,nu2
          zz(1:length,n) = Kernel_y(k+n-1)
        end do

        yy(1:sz) = 0d0

        do n = 1, int(ParKer(6))
          mu = n
          c(1:length) = 1d0
          do m = 1,nu2
            if (m .ne. mu) c(1:length) = c(1:length)*((d(1:length)+nu-m)/(mu-m))
          end do
          yy(1:length) = yy(1:length) + c(1:length)*zz(1:length,mu)

          mu = nu2+1-n
          c(1:length) = 1d0
          do m = 1,nu2
            if (m .ne. mu) c(1:length) = c(1:length)*((d(1:length)+nu-m)/(mu-m))
          end do
          yy(1:length) = yy(1:length) + c(1:length)*zz(1:length,mu)
        end do
        deallocate(c)
        deallocate(zz)
    end select

    !y(:) = unpack(yy(:), ind, y(:))
    j = 1
    do i = 1,sizey
      if(ind(i)) then
        y(i) = yy(j)
        j = j + 1
      end if
    end do

    deallocate(yy)
    deallocate(d)
    deallocate(k)
    deallocate(x0)
    deallocate(ind)

  end subroutine kernel_values

  !LAGRANGE_BASIS computes in  b  the values of Lagrange basis functions 
  ! for  m  integer knots  at the point  x.
  ! If m=2*k the knots are -k+1:k and 0<=x<=1
  ! If m=2*k+1 the knots are -k:k and -1/2<=x<=1/2

  ! Version: May 2014
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>
  function lagrange_basis(m,x) result(b)
    integer, intent(in) :: m
    double precision, dimension(:), intent(in) :: x
    double precision, dimension(size(x),m) :: b

    b(:,:)=0d0

    select case(m)
      case(10)
        b(:,1)=-(x+3)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-4)*(x-5)/362880d0
        b(:,2)=(x+4)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-4)*(x-5)/40320d0
        b(:,3)=-(x+4)*(x+3)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-4)*(x-5)/10080d0
        b(:,4)=(x+4)*(x+3)*(x+2)*x*(x-1)*(x-2)*(x-3)*(x-4)*(x-5)/4320d0
        b(:,5)=-(x+4)*(x+3)*(x+2)*(x+1)*(x-1)*(x-2)*(x-3)*(x-4)*(x-5)/2880d0
        b(:,6)=(x+4)*(x+3)*(x+2)*(x+1)*x*(x-2)*(x-3)*(x-4)*(x-5)/2880d0
        b(:,7)=-(x+4)*(x+3)*(x+2)*(x+1)*x*(x-1)*(x-3)*(x-4)*(x-5)/4320d0
        b(:,8)=(x+4)*(x+3)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-4)*(x-5)/10080d0
        b(:,9)=-(x+4)*(x+3)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-5)/40320d0
        b(:,10)=(x+4)*(x+3)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-4)/362880d0
      case(8)
        b(:,1)=-(x+2)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-4)/5040d0
        b(:,2)=(x+3)*(x+1)*x*(x-1)*(x-2)*(x-3)*(x-4)/720d0
        b(:,3)=-(x+3)*(x+2)*x*(x-1)*(x-2)*(x-3)*(x-4)/240d0
        b(:,4)=(x+3)*(x+2)*(x+1)*(x-1)*(x-2)*(x-3)*(x-4)/144d0
        b(:,5)=-(x+3)*(x+2)*(x+1)*x*(x-2)*(x-3)*(x-4)/144d0
        b(:,6)=(x+3)*(x+2)*(x+1)*x*(x-1)*(x-3)*(x-4)/240d0
        b(:,7)=-(x+3)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-4)/720d0
        b(:,8)=(x+3)*(x+2)*(x+1)*x*(x-1)*(x-2)*(x-3)/5040d0
      case(6)
        b(:,1)=-(x+1)*x*(x-1)*(x-2)*(x-3)/120d0
        b(:,2)=(x+2)*x*(x-1)*(x-2)*(x-3)/24d0
        b(:,3)=-(x+2)*(x+1)*(x-1)*(x-2)*(x-3)/12d0
        b(:,4)=(x+2)*(x+1)*x*(x-2)*(x-3)/12d0
        b(:,5)=-(x+2)*(x+1)*x*(x-1)*(x-3)/24d0
        b(:,6)=(x+2)*(x+1)*x*(x-1)*(x-2)/120d0
      case(5)
        b(:,1)=(x+1)*x*(x-1)*(x-2)/24d0
        b(:,2)=-(x+2)*x*(x-1)*(x-2)/6d0
        b(:,3)=(x+2)*(x+1)*(x-1)*(x-2)/4d0
        b(:,4)=-(x+2)*(x+1)*x*(x-2)/6d0
        b(:,5)=(x+2)*(x+1)*x*(x-1)/24d0
      case(4)
        b(:,1)=-x*(x-1)*(x-2)/6d0
        b(:,2)=(x+1)*(x-1)*(x-2)/2d0
        b(:,3)=-(x+1)*x*(x-2)/2d0
        b(:,4)=(x+1)*x*(x-1)/6d0
      case(3)
        b(:,1)=x*(x-1)/2d0
        b(:,2)=-(x+1)*(x-1)
        b(:,3)=(x+1)*x/2d0
      case(2)
        b(:,1)=1d0-x
        b(:,2)=x
      case(1)
        b(:,1)=1d0
    end select

  end function lagrange_basis


  !      subroutine gd2gc(phi,ht,ae,b,phig,re)
  !     CONVERTS COORDINATES FROM GEODETIC (phi,ht) TO GEOCENTRIC (phig,re)
  !
  ! Remark. phi and phig are geograpic coordinates, i.e. angular distances to
  ! the equator (not to the notrh pole)!!!

  !       ae = SEMI-MAJOR AXIS .....................(m)
  !       b  = SEMI-MINOR AXIS .....................(m)
  !function [phig,re]=gd2gc(phi,ht,ae,b)
  subroutine gd2gc(phi,ht,ae,b,phig,re)
    double precision, dimension(:), intent(in) :: phi, ht
    double precision, intent(in) :: ae, b
    double precision, dimension(:), intent(out) :: phig, re

    integer :: sz
    double precision :: e2
    double precision, dimension(:), allocatable :: u, t, en, rho, z

    if (size(phi).ne.size(ht)) then
      print *,'GC2GE:phi_ht'
      print *,'phi and ht must be of the same size'
      stop
    end if
    if (size(phi).ne.size(phig)) then
      print *,'GC2GE:phi_phig'
      print *,'phi and phig must be of the same size'
      stop
    end if
    if (size(phi).ne.size(re)) then
      print *,'GC2GE:phi_re'
      print *,'phi and re must be of the same size'
      stop
    end if

    sz = size(phi)

    allocate(u(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: u'
      stop
    end if
    allocate(t(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: t'
      stop
    end if
    allocate(en(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: en'
      stop
    end if
    allocate(rho(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: rho'
      stop
    end if
    allocate(z(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: z'
      stop
    end if

    e2 = (ae**2-b**2)/(ae**2)
    u = dsin(phi)
    t = dcos(phi)
    en = ae/dsqrt(1d0 - e2*u**2)
    rho = (en + ht)*t
    z = (en*(1d0-e2) + ht)*u
    phig = datan2(z,rho)
    re = dsqrt(rho**2 + z**2)

    deallocate(z)
    deallocate(rho)
    deallocate(en)
    deallocate(t)
    deallocate(u)

  end subroutine gd2gc

  !----------------------------------------------------------------------------
  ! dlmsize: Count the lines in the dlm file
  integer function dlmsize(fname)

    character*(*), intent(in) :: fname
    integer n, ret_i

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read points file.'
      stop
    end if

    n = 0
    do
      read(1, *, iostat=ret_i) 
      if(ret_i .ne. 0) exit
      n = n + 1
    end do
    close(1)

    dlmsize = n
  end function dlmsize

  !----------------------------------------------------------------------------
  !  dlmread:   Reads scattered point coordinates from free-format ASCII file
  !     replaces the matlab function DLMREAD
  !     FORTRAN VERSION:                          L. Scott Johnson,  APR 2013
  subroutine dlmread(fname, n, ptarr)

    character*(*), intent(in) :: fname
    integer, intent(in) :: n
    double precision :: ptarr(n,3)
    integer :: i, ret_i

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read points file ', fname
      stop
    end if

    !--------------------------------------------------------------------------
    !     read the coordinates from the file into the array
    !--------------------------------------------------------------------------
    do i=1,n
      read(1,*) ptarr(i,1), ptarr(i,2), ptarr(i,3)
    end do
    close(1)
  end subroutine dlmread

end program emmsynth_fast
