program emmsynth_init
  !EMMSYNT_INIT evaluates the weighted regular grid values
  ! on multiple confocal ellipsoids of the magnetic field gradient
  ! and writes them in an unformatted data FORTRAN file 's3'.
  ! The gradient is defined via the magnetic model
  ! coefficient files:
  ! 'EMM2015.COF' and 'EMM2015SV.COF' for Enhanced Magnetic Model EMM2015 or
  ! 'EMM2017.COF' and 'EMM2017SV.COF' for Enhanced Magnetic Model EMM2017.
  !
  ! Version: November 2018
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  integer :: model;
  integer :: modeldegree;
  character(len=80) :: in1, in2, out1, out2, out3, out4, out5, out6
  integer*8 :: icount, icount_rate
  real*8 :: time0, time1, time2, time3, time4, time5, time6
  real*8 :: TTime0, TTime1, TTime2, TTime3, TTime4
  integer :: terms_small, db, NN, NumHeights
  integer :: KK, LL, LL2, KKext, LLext
  integer :: i, k, k1, k2, mm, m, nu, ret_i
  integer :: ndelta1, ndelta2
  double precision :: h, tau, ae, inv_flat, be, E, b0
  double precision :: weight, chi, u, pi
  double precision :: am, arat, tmp, tmp1
  double precision, dimension(:), allocatable :: v, heights, phi, lambda
  double precision, dimension(:), allocatable :: r1, w
  double precision, dimension(:,:), allocatable :: lambda0
  double precision, dimension(:,:), allocatable :: cnm, snm, cnmw, snmw
  double precision, dimension(:,:), allocatable :: cnm1, snm1, cnme1, snme1
  double precision, dimension(:,:), allocatable :: cnm2, snm2, cnme2, snme2
  double precision, dimension(:,:), allocatable :: valueXext, valueX
  character(len=3) :: tip
  character(len=3), dimension(6) :: tipp
  integer :: tipi
  integer :: nmax, nmax1, nmaxe, nmaxe1, n
  character(len=80) :: s1,s3

  parameter (pi = 3.14159265358979323846264338327950288d0)
  
  TTime1=0d0
  TTime2=0d0
  TTime3=0d0
  TTime4=0d0

  ! model and year input and check
  write(*, fmt='(A)', advance="no") &
    'Which model to use? (Enter 1 for EMM2015 or 2 for EMM2017): '
  read (*,*) model

  select case(model)
    case(1)
      modeldegree=740
      in1='EMM2015.COF'
      in2='EMM2015SV.COF'
      out1='emm2015_Xp.bin'
      out2='emm2015_Yp.bin'
      out3='emm2015_Zp.bin'
      out4='emm2015_XpSV.bin'
      out5='emm2015_YpSV.bin'
      out6='emm2015_ZpSV.bin'
    case(2)
      modeldegree=790
      in1='EMM2017.COF'
      in2='EMM2017SV.COF'
      out1='emm2017_Xp.bin'
      out2='emm2017_Yp.bin'
      out3='emm2017_Zp.bin'
      out4='emm2017_XpSV.bin'
      out5='emm2017_YpSV.bin'
      out6='emm2017_ZpSV.bin'
    case default
      print *,'emmsynth_init:model','wrong input'
      stop
  end select

  call system_clock(icount, icount_rate)
  time0 = dble(icount)/dble(icount_rate)

  ! Earth constants
  ! wgs84_constants
  ae = 6378137d0 !m
  inv_flat=298.257223563d0
  be = ae*(1d0-1d0/inv_flat)
  E=dsqrt((ae-be)*(ae+be))

  am=6371200 !m; geomagnetic reference radius

  ! displacement for the lowest Earth surface point wrt reference ellipsoid
  db=-417         !reference ellipsoid displacement in meters $b_0=b_e-417$
  b0=be+dble(db)

  terms_small=40

  ! nodal set
  NN=740; !parameter for the nodal set

  tau=2d0
  ndelta1=9
  ndelta2=9

  ! regular grid dimensions
  KK=int((1d0+tau/2d0)*dble(NN) + 0.5d0) + 1
  LL2=int((1d0+tau/2d0)*dble(NN) + 0.5d0)
  LL=2*LL2

  weight=1d0/dble((KK-1)*LL2)

  allocate(phi(KK), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: phi'
    stop
  end if
  phi(1:KK) = (/ (dble(i)*pi/dble(KK-1), i=0,KK-1) /)

  allocate(lambda(LL), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: lambda'
    stop
  end if
  lambda(1:LL) = (/ (dble(i)*2d0*pi/dble(LL), i=0,LL-1) /)

  allocate(lambda0(KK,LL), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: lambda0'
    stop
  end if
  do k=1,KK
    lambda0(k,1:LL) = lambda(1:LL)
  end do

  ! extended values parameters

  ! extended dimensions
  KKext=KK+2*ndelta1-1
  LLext=LL+2*ndelta2-1

  ! generating heights
  h=2.0124185d-2
  NumHeights=45

  allocate(v(NumHeights), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: v'
    stop
  end if
  v(1:NumHeights) = (/ (1d0 - (dble(i)*h)**4 / 4d0, i=0,NumHeights-1) /)

  allocate(heights(NumHeights), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: heights'
    stop
  end if
  heights(1:NumHeights) = b0*(1d0/v(1:NumHeights)-1d0);  ! heights in meters

  deallocate(v)

  print *, 'number of rows   number of columns   number of ellipsoids'
  print *, KKext, LLext, NumHeights

  ! evaluation of geocentric magnetic quantities

  tipp = (/ 'X  ', 'XSV', 'Y  ', 'YSV', 'Z  ', 'ZSV' /)

  do tipi = 1, size(tipp)
    tip = tipp(tipi)
    select case(tip)
      case('X  ')
        nmax=modeldegree
        s1=in1
        s3=out1
        chi= -1d0
      case('XSV')
        nmax=15
        s1=in2
        s3=out4
        chi= -1d0
      case('Y  ')
        nmax=modeldegree
        s1=in1
        s3=out2
        chi= -1d0
      case('YSV')
        nmax=15
        s1=in2
        s3=out5
        chi= -1d0
      case('Z  ')
        nmax=modeldegree
        s1=in1
        s3=out3
        chi=1d0
      case('ZSV')
        nmax=15
        s1=in2
        s3=out6
        chi=1d0
      case default
        print *, 'emmsynth_init:tip','wrong value of "tip"'
        stop
    end select

    nmax1=nmax+1
    nmaxe=nmax+30
    nmaxe1=nmaxe+1

    ! Ellipsoidal harmonic series coefficients

    ! Read spherical harmonic geomagnetic model
    call system_clock(icount)
    time1=dble(icount)/dble(icount_rate)

    allocate(cnm(nmax1,nmax1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: cnm'
      stop
    end if
    allocate(snm(nmax1,nmax1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: snm'
      stop
    end if

    cnm(1:nmax1,1:nmax1) = 0d0
    snm(1:nmax1,1:nmax1) = 0d0

    call coeff_read(s1, cnm, snm)
    
    call system_clock(icount)
    time2=dble(icount)/dble(icount_rate)
    TTime1 = TTime1 + time2-time1

    ! Modify spherical harmonic geomagnetic model

    ! from Schmidt semi-normalized ALF to fully-normalized ALF

    do n=1,nmax1
      tmp = ((2d0*dble(n-1) + 1d0)**(-0.5d0))
      cnm(n,:) = cnm(n,:) * tmp
      snm(n,:) = snm(n,:) * tmp
    end do

    !from reference sphere with radius 'am' to reference sphere with radius 'ae'
    arat = am/ae

    tmp = 1d0
    do n = 1, nmax1
      tmp=tmp*arat
      cnm(n,:)=cnm(n,:)*tmp
      snm(n,:)=snm(n,:)*tmp
    end do


    allocate(cnm1(nmax1,nmax1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: cnm1'
      stop
    end if
    allocate(snm1(nmax1,nmax1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: snm1'
      stop
    end if

    ! Spherical coefficient formation

    select case(tip)

      case('X  ','XSV')
        
        allocate(cnm2(nmax1,nmax1), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: cnm2'
          stop
        end if
        allocate(snm2(nmax1,nmax1), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: snm2'
          stop
        end if

        cnm1 = 0
        snm1 = 0
        cnm2 = 0
        snm2 = 0

        tmp=1d0

        allocate(w(nmax1), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: w'
          stop
        end if

        do n=2,2
          w = (/ ( (dsqrt(dble(n-mm)*(n+mm-1)))/2d0, mm=1,n ) /)
          w(1)=w(1)*dsqrt(2d0)
          cnm1(n,1)=w(1)*cnm(n,2)*tmp
          cnm1(n,n)=-w(n-1)*cnm(n,n-1)*tmp
          cnm2(n,1)=w(1)*snm(n,2)*tmp
          snm2(n,n)=-w(n-1)*cnm(n,n-1)*tmp
        end do
        do n=3,nmax1
          snm(n,1)=0d0 !just in case this is not fulfilled
          w = (/ ( (dsqrt(dble(n-mm)*(n+mm-1)))/2d0, mm=1,n ) /)
          w(1)=w(1)*dsqrt(2d0)
          cnm1(n,1)=w(1)*cnm(n,2)*tmp
          cnm1(n,2:n-1)=(w(2:n-1)*cnm(n,3:n)-w(1:n-2)*cnm(n,1:n-2))*tmp
          cnm1(n,n)=-w(n-1)*cnm(n,n-1)*tmp
          snm1(n,2:n-1)=(w(2:n-1)*snm(n,3:n)-w(1:n-2)*snm(n,1:n-2))*tmp
          snm1(n,n)=-w(n-1)*snm(n,n-1)*tmp
          cnm2(n,1)=w(1)*snm(n,2)*tmp
          cnm2(n,2:n-1)=(w(2:n-1)*snm(n,3:n)+w(1:n-2)*snm(n,1:n-2))*tmp
          cnm2(n,n)=w(n-1)*snm(n,n-1)*tmp
          snm2(n,2:n-1)=(-w(2:n-1)*cnm(n,3:n)-w(1:n-2)*cnm(n,1:n-2))*tmp
          snm2(n,n)=-w(n-1)*cnm(n,n-1)*tmp
        end do

        deallocate(w)

      case('Y  ', 'YSV')
        allocate(cnm2(nmax1,nmax1), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: cnm2'
          stop
        end if
        allocate(snm2(nmax1,nmax1), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: snm2'
          stop
        end if

        cnm1 = 0
        snm1 = 0
        cnm2 = 0
        snm2 = 0
        tmp=0.5d0/arat

        allocate(w(2*nmax1), stat=ret_i)
        if(ret_i .gt. 0) then
          print *, 'Out of memory: w'
          stop
        end if
        w = (/ (dsqrt(dble(i-1)*dble(i)), i=1,2*nmax1) /)

        cnm1(1:nmax,1) = -dsqrt(2d0)*w(2:nmax1)*snm(2:nmax1,2)
        cnm2(1:nmax,1) = dsqrt(2d0)*w(2:nmax1)*cnm(2:nmax1,2)

        do m=2,nmax
          cnm1(m:nmax,m) = -w(m+m:m+nmax)*snm(m+1:nmax1,m+1)
          snm1(m:nmax,m) = w(m+m:m+nmax)*cnm(m+1:nmax1,m+1)
          cnm2(m:nmax,m) = w(m+m:m+nmax)*cnm(m+1:nmax1,m+1)
          snm2(m:nmax,m) = w(m+m:m+nmax)*snm(m+1:nmax1,m+1)
        end do

        do m=3,nmax
          cnm1(m:nmax,m) = cnm1(m:nmax,m)-w(2:nmax-m+2)*snm(m+1:nmax1,m-1)
          snm1(m:nmax,m) = snm1(m:nmax,m)+w(2:nmax-m+2)*cnm(m+1:nmax1,m-1)
          cnm2(m:nmax,m) = cnm2(m:nmax,m)-w(2:nmax-m+2)*cnm(m+1:nmax1,m-1)
          snm2(m:nmax,m) = snm2(m:nmax,m)-w(2:nmax-m+2)*snm(m+1:nmax1,m-1)
        end do

        deallocate(w)

        do m=1,nmax1
          tmp1 = dsqrt(dble(2*m+1)/dble(2*m-1))*tmp
          cnm1(m,:) = cnm1(m,:)*tmp1
          cnm2(m,:) = cnm2(m,:)*tmp1
          snm1(m,:) = snm1(m,:)*tmp1
          snm2(m,:) = snm2(m,:)*tmp1
        end do

      case('Z  ', 'ZSV')
        
        do m=1,nmax1
          cnm1(m,:) = -cnm(m,:)*m
          snm1(m,:) = -snm(m,:)*m
        end do

      case default
        print *, 'Invalid value for tip: ', tip
        stop

    end select

    deallocate(cnm)
    deallocate(snm)

    ! Ellipsoidal coefficient formation

    allocate(cnmw(nmax+2*terms_small+1,nmax+2*terms_small+1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: cnmw'
      stop
    end if
    allocate(snmw(nmax+2*terms_small+1,nmax+2*terms_small+1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: snmw'
      stop
    end if

    call jekeli_sphere2ellipsoid_coeff(cnm1,snm1,nmax,be,E,ae,terms_small, &
        cnmw, snmw)

    deallocate(snm1)
    deallocate(cnm1)

    allocate(cnme1(nmaxe1,nmaxe1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: cnme1'
      stop
    end if
    allocate(snme1(nmaxe1, nmaxe1), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: snme1'
      stop
    end if

    cnme1(1:nmaxe1,1:nmaxe1) = cnmw(1:nmaxe1,1:nmaxe1)
    snme1(1:nmaxe1,1:nmaxe1) = snmw(1:nmaxe1,1:nmaxe1)

    if((tip.eq.'X  ').or.(tip.eq.'XSV').or.(tip.eq.'Y  ').or.(tip.eq.'YSV')) &
      then
      call jekeli_sphere2ellipsoid_coeff(cnm2,snm2,nmax,be,E,ae,terms_small, &
          cnmw, snmw)

      deallocate(cnm2)
      deallocate(snm2)

      allocate(cnme2(nmaxe1,nmaxe1), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: cnme2'
        stop
      end if
      allocate(snme2(nmaxe1,nmaxe1), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: snme2'
        stop
      end if

      cnme2=cnmw(1:nmaxe1,1:nmaxe1)
      snme2=snmw(1:nmaxe1,1:nmaxe1)

    end if

    deallocate(cnmw)
    deallocate(snmw)

    call system_clock(icount)
    time3=dble(icount)/dble(icount_rate)
    TTime2 = TTime2 + time3-time2

    ! EHS values

    allocate(valueXext(KKext,LLext), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: valueXext'
      stop
    end if

    valueXext = 0d0

    ! open file  s3  for writing in machine format:IEEE floating point
    ! with little-endian byte ordering and 64-bit long data type
    open(10,file=s3,form='unformatted',action='write',status='replace', &
        iostat=ret_i)

    if(ret_i .gt. 0) then
      print *, 'error opening ', s3, ' for writing.'
      stop
    end if

    print *, tip

    allocate(valueX(KK,LL), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: valueX'
      stop
    end if
    allocate(r1(KK), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: r1'
      stop
    end if

    do nu=1,size(heights)
      call system_clock(icount)
      time3=dble(icount)/dble(icount_rate)

      u=b0+heights(nu)

      r1 = ellips2sphere_r(phi,0d0,u,E) !1

      valueX = ellipsoid_harmonic_series_grid(phi,LL,u,be,E,cnme1,snme1,1d260)

      if((tip.eq.'X  ').or.(tip.eq.'XSV').or.(tip.eq.'Y  ').or.(tip.eq.'YSV')) &
        then
        valueX =valueX * dcos(lambda0) + &
          ellipsoid_harmonic_series_grid(phi,LL,u,be,E,cnme2,snme2,1d260) * &
          dsin(lambda0)
      end if

      if ((tip.eq.'Y  ').or.(tip.eq.'YSV')) then
        do i=1,LL
          valueX(:,i) = valueX(:,i) * ((am/r1)**2) * weight
        end do
      else
        do i=1,LL
          valueX(:,i) = valueX(:,i) * (am/r1) * weight
        end do
      end if

    !   Spherical extension for  ParCub(1)==1
      valueXext(ndelta1:ndelta1+KK-1, ndelta2:ndelta2+LL-1) = valueX

      valueXext(1:ndelta1-1, ndelta2:ndelta2+LL2-1) = &
        chi*valueX(ndelta1:2:-1, LL2+1:LL)

      valueXext(1:ndelta1-1, ndelta2+LL2:ndelta2+LL-1) = &
        chi*valueX(ndelta1:2:-1, 1:LL2)

      valueXext(ndelta1+KK:KK+2*ndelta1-1, ndelta2:ndelta2+LL2-1) = &
        chi*valueX(KK-1:KK-ndelta1:-1, LL2+1:LL)

      valueXext(ndelta1+KK:KK+2*ndelta1-1, ndelta2+LL2:ndelta2+LL-1) = &
        chi*valueX(KK-1:KK-ndelta1:-1, 1:LL2)

      valueXext(:, 1:ndelta2-1) = valueXext(:, LL+1:LL+ndelta2-1)

      valueXext(:, LL+ndelta2:LL+2*ndelta2-1) = &
        valueXext(:, ndelta2:2*ndelta2-1)

      call system_clock(icount)
      time4=dble(icount)/dble(icount_rate)
      TTime3 = TTime3 + time4 - time3

      !fwrite(fid1, KKext*LLext*8, 'uint32')
      !fwrite(fid1, valueXext, 'double')
      !fwrite(fid1, KKext*LLext*8, 'uint32')
      write(10) ((valuexext(k1,k2), k1=1,kkext), k2=1,llext)

      call system_clock(icount)
      time5=dble(icount)/dble(icount_rate)
      TTime4 = TTime4 + time5 - time4

      print *, nu, heights(nu), maxval(abs(valueX))/weight

    end do ! nu = 1,size(heights)

    deallocate(r1)
    deallocate(valueX)
    deallocate(valueXext)

    if((tip.eq.'X  ').or.(tip.eq.'XSV').or.(tip.eq.'Y  ').or.(tip.eq.'YSV')) &
      then
      deallocate(snme2)
      deallocate(cnme2)
    end if

    deallocate(snme1)
    deallocate(cnme1)

    close(10)

  end do ! i = 1, size(tipp) for tipp=['X' 'XSV' 'Y' 'YSV' 'Z' 'ZSV']

  deallocate(lambda0)
  deallocate(lambda)
  deallocate(phi)
  deallocate(heights)

  call system_clock(icount)
  time6=dble(icount)/dble(icount_rate)
  TTime0 = time6 - time0

  ! statistics

  write(*,'(A)') &
    '---------------------------------------------------------------'
  select case(model)
    case(1)
      write(*,'(A)') '        Program:  emmsynth_init,       Model: EMM2015'
    case(2)
      write(*,'(A)') '        Program:  emmsynth_init,       Model: EMM2017'
  end select
  write(*,'(A)') &
    '---------------------------------------------------------------'
  write(*,'(A,F12.5,A)') &
    '  Total time                  = ', TTime0, ' CPU seconds'
  write(*,'(A,F12.5,A)') &
    '  Coefficients load time      = ', TTime1, ' CPU seconds'
  write(*,'(A,F12.5,A)') &
    '  Coefficients formation time = ', TTime2, ' CPU seconds'
  write(*,'(A,F12.5,A)') &
    '  Synthesis time              = ', TTime3, ' CPU seconds'
  write(*,'(A,F12.5,A)') &
    '  Values write time           = ', TTime4, ' CPU seconds'
  write(*,'(A)') &
    '---------------------------------------------------------------'

      contains

  !JEKELI_SPHERE2ELLIPSOID_COEFF computes the ellipsoidal harmonic coefficients of
  ! a harmonic function given by its spherical harmonic coefficients.
  ! The coefficients are computed via Jekeli transform for the approximate identity
  ! $\sum_{n=0}^N \sum_{m=0}^n (a/r)^{n+1}
  ! (cnmS_{nm}\cos(m\lambda)+snmS_{nm}\sin(m\lambda)\tilde{P}_{nm}(\theta)
  ! = \sum_{n=0}^{N+2 terms} \sum_{m=0}^n \tilde{Q}_{nm}(iu/E)/\tilde{Q}_{nm}(ib/E) 
  ! (cnmE_{nm}\cos(m\lambda)+snmE_{nm}\sin(m\lambda)\tilde{P}_{nm}(\phi)$,
  ! where (theta,lambda,r) and (phi,lambda,u) are the spherical and ellipsoidal
  ! harmonic coordinates of a point in $R^3$.
  !
  ! INPUT
  ! 1. 'cnmS', 'snmS'  are lower triangular matrices with N+1 rows holding
  !    the cosine and sine spherical coefficients.
  ! 2. 'N' is the degree of the spherical harmonic series.
  ! 3. 'b' and 'E' are semi-minor axis and the linear eccentricity of the
  !    reference ellipsoid.
  ! 4. 'R' is the radius of the reference sphere.
  ! 5. 'terms' is a parameter for function 'lambda_small'.
  !
  ! OUTPUT
  ! 1. 'cnmE', 'snmE'  are lower triangular matrices with N+2*terms+1 rows
  !    holding the cosine and sine ellipsoidal harmonic coefficients.
  !
  ! Notes.
  ! 1. The code requires functions 'lambda_small' and 'hypergeometric21_alf4'.
  ! 2. Parameter '18' in 'hypergeometric21_alf4' is determined for the Earth
  !    ellipsoid.

  ! Version: October 2015
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  subroutine jekeli_sphere2ellipsoid_coeff(cnmS,snmS,N,b,E,R,terms,cnmE,snmE)
    double precision, dimension(N+1,N+1), intent(in) :: cnmS, snmS
    integer, intent(in) :: N
    double precision, intent(in) :: b, E, R
    integer, intent(in) :: terms
    double precision, dimension(N+2*terms+1,N+2*terms+1), intent(out) :: &
      cnmE, snmE

    integer :: m, nu, i, w
    double precision :: z2
    double precision, dimension(N+2*terms+1) :: S
    double precision, dimension(N+2*terms+1) :: coefSc, coefSs, coefEc, coefEs
    double precision, dimension(terms) :: x1, x2

    cnmE=0d0
    snmE=0d0
    z2=(E/b)**2

    do m=0,N

      S = (1d0+z2)**(m/2d0) * (/ ((R/b)**(nu), nu=1,N+2*terms+1) /) * &
          hypergeometric21_alf4(N+2*terms,m,-z2,18)

      coefSc(1:N+2*terms+1)=0d0
      coefSs(1:N+2*terms+1)=0d0
      coefEc(1:N+2*terms+1)=0d0
      coefEs(1:N+2*terms+1)=0d0

      coefSc(1:N+1)=cnmS(:,m+1)
      coefSs(1:N+1)=snmS(:,m+1)

      do i=m,N+2*terms
        w=min(int(floor(dble(i-m)/2d0)),terms-1)
        x1=lambda_small(i,m,E/R,terms)
        x2(1:w+1) = S(i+1)*x1(1:w+1)

        coefEc(i+1) = dot_product(x2(1:w+1),coefSc(i+1:i+1-2*w:-2))
        coefEs(i+1) = dot_product(x2(1:w+1),coefSs(i+1:i+1-2*w:-2))

      end do

      cnmE(:,m+1)=coefEc
      snmE(:,m+1)=coefEs

    end do ! m=0,N
  end subroutine jekeli_sphere2ellipsoid_coeff

  !HYPERGEOMETRIC21_ALF4 computes the HG series for ALF of the second kind 
  ! 2F1((n+m+1)/2,(n+m+2)/2;(2n+3)/2;z)
  ! using linear and quadratic transformations.
  ! 'N' defines the array n=(0:N)' of spherical degrees. max(max(abs(z)))<1.
  !
  ! Use degree=18 for geopotential studies with N<=2300.

  ! Version: October 2015
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function hypergeometric21_alf4(N,m_arg,z,degree) result(y)
    integer, intent(in) :: N, m_arg
    double precision, intent(in) :: z
    integer, intent(in) :: degree
    double precision, dimension(N+1) :: y

    integer :: sz, m
    integer, dimension(:), allocatable :: n_
    double precision :: g, z1, a, b, z2, zlog
    double precision, dimension(:), allocatable :: alpha, beta, gamma_
    double precision, dimension(:), allocatable :: temp, wy
    double precision, dimension(:), allocatable :: gv

    m = m_arg
    if (m .lt. 0) m = -m

    if (N .lt. 0) then
      print *, 'hypergeometric21_alf4:N'
      print *, '"N" must be non-negative'
      stop
    end if

    y=0d0

    if (N .lt. m) return

    if (m .ne. 0) then
      sz = min(2*m-1,N) - m + 1
      allocate(n_(sz), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: n_'
        stop
      end if
      allocate(alpha(sz), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: alpha'
        stop
      end if
      allocate(beta(sz), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: beta'
        stop
      end if
      allocate(gamma_(sz), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: gamma_'
        stop
      end if
      allocate(temp(sz), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: temp'
        stop
      end if
      allocate(wy(sz), stat=ret_i)
      if(ret_i .gt. 0) then
        print *, 'Out of memory: wy'
        stop
      end if

      n_(1:sz) = (/ (i, i=m,min(2*m-1,N)) /)
      alpha = 0.5d0*dble(n_-m+1)
      beta = alpha+.5d0
      gamma_ = dble(n_)+1.5d0
      g = dexp(-log1p(-z)*dble(m))
      z1=z
      temp(1:sz)=1d0
      wy(1:sz)=1d0
      do k=0,degree-1
        temp = temp*(alpha+dble(k))*(beta+dble(k))/(gamma_+dble(k)) &
          *(z1/dble(k+1))
        wy = wy+temp
      end do
      y(n_+1)=g*wy

      deallocate(wy)
      deallocate(temp)
      deallocate(gamma_)
      deallocate(beta)
      deallocate(alpha)
      deallocate(n_)

    end if 

    if (N .lt. 2*m) return

    sz = N - 2*m + 1
    allocate(n_(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: n_'
      stop
    end if
    allocate(gamma_(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: gamma_'
      stop
    end if
    allocate(temp(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: temp'
      stop
    end if
    allocate(gv(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: gv'
      stop
    end if
    allocate(wy(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: wy'
      stop
    end if
    n_(1:sz) = (/ (i, i=2*m,N) /)
    a = dble(m)+.5d0
    b = dble(-m)+.5d0
    gamma_(1:sz) = dble(n_(1:sz))+1.5d0
    zlog = .5 * log1p(-z)
    z2 = -expm1(zlog) 
    gv(1:sz) = dexp(-zlog*(dble(m)+.5)-log1p(-z2/2d0)*(dble(n_(1:sz))+.5))
    z1 = -z/(4d0-2d0*(z2+z))
    temp(1:sz) = 1d0
    wy = temp
    do k=0,degree-1
      temp = temp/(gamma_+dble(k))*(z1*(a+dble(k))*(b+dble(k))/dble(k+1))
      wy = wy + temp
    end do
    y(n_+1)=gv*wy;

    deallocate(wy)
    deallocate(gv)
    deallocate(temp)
    deallocate(gamma_)
    deallocate(n_)

  end function hypergeometric21_alf4

  !LAMBDA_SMALL is the $\lambda$ function of Jekeli transform, 1988,
  ! where $y(k+1)=\lambda_{n,m,k}$, x=E/a.
  ! 
  ! INPUT
  ! 1. 'n', 'm'  are integers satisfying 0 <= m <= n.
  ! 2. 'x' is real, x=E/a in the case of Earth reference sphere and ellipsoid
  ! 5. 'terms' is the lenth of the coefficient vector 'y'
  !
  ! OUTPUT
  ! 1. 'y' is the coefficient vector of length 'terms'
  !
  ! Notes.
  ! terms=40 cuts  lambda  at 10^{-25} for every 0<=m<=n<=2190 and x=E/a.

  ! Version: October 2015
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function lambda_small(n,m,x,terms) result(y)
    integer, intent(in) :: n, m
    double precision, intent(in) :: x
    integer, intent(in) :: terms

    double precision, dimension(terms) :: y

    integer :: w, w1, k
    double precision :: tmp

    y=0d0

    if (abs(m) .gt. n) return

    y(1)=1d0
    w =floor(dble(n-abs(m))/2d0)
    w1=min(w,terms-1)
    do k=1,w1
      tmp=dsqrt((dble(n-m-2*k+1)*dble(n-m-2*k+2)*dble(n+m-2*k+1)*dble(n+m-2*k+2)*dble(2*n-4*k+1)&
          / dble(2*n-4*k+5)))
      y(k+1)=tmp/dble(2*k*(2*n-2*k+1))*(x**2)*y(k)
    end do

  end function lambda_small

  !ELLIPS2SPHERE converts the ellipsoidal coordinates (phi,lambda,u)
  ! of a 3-d point spherical coordinates (theta,lambda,r).
  ! 'e' is the linear eccentricity of the confocal ellipses.
  ! e^2=a^2-b^2, where 'a', 'b' are the semi-axes.
  ! Forward stability for u>e.
  ! Requirements: The arrays 'phi', 'lambda', 'u' must be of the same size;
  ! '0<=phi<=pi', '0<=lambda<2 pi', 'u>=0'; 'e' must be non-negative.
  !
  ! FORTRAN Version, ellips2sphere_r, returns only the r values, and
  ! takes scalar lambda and scalar u and scalar e. -- LSJ

  ! Version: October 2015
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function ellips2sphere_r(phi,lambda,u,e) result(r)
    double precision, dimension(:), intent (in) :: phi
    double precision, intent(in) :: lambda, u, e
    double precision, dimension(size(phi)) :: r

    if(e .lt. 0) then
      print *, 'ellips2sphere:e'
      print *, '"e" must be non-negative'
      stop
    end if

    if(any(phi .gt. pi) .or. any(phi .lt. 0d0)) then
      print *, 'ellips2sphere:phi'
      print *, '"phi" must be between 0 and pi'
      stop
    end if

    if(lambda .ge. 2d0*pi .or. lambda .lt. 0d0) then
      print *, 'ellips2sphere:lambda'
      print *, '"lambda" must be between 0 and 2*pi'
      stop
    end if

    if(u .lt. 0d0) then
      print *, 'ellips2sphere:u'
      print *, '"u" must be non-negative'
      stop
    end if

    r=dsqrt(u**2 + (e**2)*dsin(phi)**2)

  end function ellips2sphere_r

  ! ELLIPSOID_HARMONIC_SERIES_GRID computes the values 
  ! of a finite ellipsoidal harmonic series at the grid
  ! with rows given by the elements of 'phi' and (0:L-1)*2*pi/L as columns
  ! on the ellipsoid with semi-minor axis 'u' and linear eccentricity 'E'.
  ! The coefficients in 'cnm', 'snm' are determined for the reference  
  ! confocal ellipsoid with semi-minor axis 'b'.
  !
  ! INPUT
  ! 1. 'phi' is a vectors with the ellipsoidal-harmonic compliment of reduced latitude
  !   coordinate of the grid points on the ellipsoid with semi-minor axis 'u'.
  ! 2. 'L', 'u', 'b', 'E' are numbers, 'L' is integer, u>=0, b>0, E>0.
  ! 3. 'cnm', 'snm'  are lower triangular matrices of the same size
  !   with the cosine and sine coefficients of the ellipsoidal harmonic series
  ! 4. 'gs' is a global scaling factor for avoiding underflow in evaluating the
  !   associated Legendre functions, say gs=10^260.
  !
  ! OUTPUT
  ! 1. 'y'  is a matrix of dimension (lenth of 'phi') x L  with the series values
  !
  ! Version: October 2015
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function ellipsoid_harmonic_series_grid(phi,L,u,b,E,cnm,snm,gs) result (y)

    double precision, dimension (:), intent(in) :: phi
    integer, intent(in) :: L
    double precision, intent(in) :: u, b, E
    double precision, dimension(:,:), intent(in) ::  cnm, snm
    double precision, intent(in) :: gs

    double precision, dimension(size(phi),L) :: y

    integer*4, dimension(1) ::  nsiz
    integer :: N1, N, K, m, k_, j
    double precision :: rho2, tau
    double precision, dimension(size(cnm,1)) :: rho1, work1, work2
    double precision, dimension(size(cnm,1),size(cnm,2)) :: rho, cnm_, snm_
    double precision, dimension(size(phi),size(cnm,1)) :: Cmt, Smt, p
    double precision, dimension(size(cnm,1),size(phi)) :: pmm
    double precision, dimension(L) :: Xc, Xs
    double precision, dimension(2*L) :: X, wrkfft
    double precision, dimension(:), allocatable :: root2

    N1=size(cnm,1)
    N=N1-1
    K=size(phi)
    nsiz(1) = L

    allocate(root2(2*N+5), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: root2'
      stop
    end if

    root2 = (/ (dsqrt(dble(i)), i=1,2*N+5) /)

    ! modification of the coefficients for the ellipsoidal harmonic case
    rho(1:N1, 1:N1) = 0d0
    rho1 = (/ ((b/u)**i, i=1,N1) /)

    rho2 = 1d0
    tau=dsqrt((1d0+(E/u)**2)/(1d0+(E/b)**2))
    do m=0,N
      work1(1:N1)=hypergeometric21_alf4(N,m,-(E/u)**2,18)
      work2(1:N1)=hypergeometric21_alf4(N,m,-(E/b)**2,18)
      rho(m+1:N1, m+1)=rho2*rho1(m+1:N1)*(work1(m+1:N1)/work2(m+1:N1))
      rho2=rho2*tau
    end do

    cnm_=rho*cnm
    snm_=rho*snm

    ! series evaluation
    y(1:K,1:L) = 0d0 !matrix with harmonic series values
    Cmt(1:K,1:N1) = 0d0
    Smt(1:K,1:N1) = 0d0

    pmm = alfsect(N,gs,phi)

    do m=1,N1
      p = alford(N,m-1,phi,pmm,root2)
      Cmt(:,m)=(matmul(p(:,:),cnm_(:,m)))
      Smt(:,m)=(matmul(p(:,:),snm_(:,m)))
    end do


    deallocate(root2)

    do k_=1,K
      Xc = pre_fft(Cmt(k_,:),L,1)
      Xs = pre_fft(Smt(k_,:),L,-1)

      do j = 1, L
        X(2*j-1) = Xc(j)
        X(2*j)   = Xs(j)
      enddo  !  j

      call fourt(X,nsiz,1,-1,1,wrkfft,L*2,L*2)
      y(k_,:) = (/ (X(2*j-1)/gs, j=1,L) /)
    end do

  end function ellipsoid_harmonic_series_grid

  !ALFSECT computes scaled point values of sectorial Associated Legendre Functions
  !
  ! INPUT
  ! 1. N is the maximum order of the required ALFs
  ! 2. gs is a global scaling factor for avoiding underflow in evaluating the
  !   Associated Legendre Functions, say gs=10^260.
  ! 3. theta is a vector with the co-latitudes
  ! 
  ! OUTPUT
  ! 1. pmm  is a matrix of size N+1 x (the lenth of  theta)  
  ! with the scaled values of sectorial ALF
  !
  ! Notes. 
  ! 1. ALFSECT follows Fortran alfpsec.f code
  ! 2. The input degree 'N' must be less than or equal to 2700.
  !
  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function alfsect(N,gs,theta) result(pmm)
    integer, intent(in) :: N
    double precision, intent(in) :: gs
    double precision, dimension(:), intent(in) :: theta
    double precision, dimension(N+1, size(theta)) :: pmm

    integer :: K, n1, small, i, nu, rn2
    integer, dimension(size(theta)) :: zero
    double precision, dimension(size(theta)) :: u
    double precision, dimension(N+1) :: c
    double precision :: gslog
    double precision :: eps
    double precision :: val

    eps = 2.2204d-16

    K=size(theta)
    zero=0

    c=0d0
    c(2) = dsqrt(3d0)
    do n1 = 3,N+1
      rn2 = (n1-1)*2;
      c(n1) = dsqrt(dble(rn2+1))/dsqrt(dble(rn2))
    end do

    small = -200
    gslog = dlog10(gs)
    u=dsin(theta)

    do i = 1,K
      if(u(i)<eps) then
        zero(i)=1
      else if (u(i)>1-eps) then
        zero(i)=N+1
      else
        val=anint((small - gslog)/dlog10(u(i)))
        if(val .gt. dble(N)) then
          zero(i) = N+1
        else
          zero(i) = int(val)+1
        end if
      end if
    end do

    pmm(1:N+1,1:K)=0d0

    do i=1,K
      pmm(1,i)=gs
      do nu=2,zero(i)
        pmm(nu,i)=c(nu)*u(i)*pmm(nu-1,i)
      end do
    end do

  end function alfsect

  !ALFORD computes scaled point values of non-sectorial Associated Legendre
  ! Functions (ALF) for a given order 'm' and for all degrees from 'm' to 'N'
  !
  ! INPUT
  ! 1. N  is the maximum degree of the required ALFs
  ! 2. m  is the order for which non-sectorial ALFs are to be computed
  ! 3. theta is a vector with the co-latitudes
  ! 4. pmm  is a matrix of size N+1 x (the lenth of  theta)  
  !   with the scaled values of sectorial ALF
  ! 5. root2 is a work vector, root2=realsqrt(1:2*N+5)'
  !
  ! OUTPUT
  ! 4. p  is a matrix of size (the lenth of  theta) x N+1
  !   with the scaled values of non-sectorial ALF
  !
  ! Notes. 
  ! 1. ALFORD follows Fortran alfpord.f code
  ! 2. The input degree 'N' must be less than or equal to 2700.
  ! 3. The input order 'm' must be less than or equal to the maximum
  !   degree 'N'. 
  ! 4. Note that any global scaling of the computed ALFs enters through the
  !   pre-computed sectorial values in 'pmm'.
  !
  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function alford(N,m,theta,pmm,root2) result(p)
    integer, intent(in) :: N, m
    double precision, dimension(:), intent(in) :: theta
    double precision, dimension(:,:), intent(in) :: pmm
    double precision, dimension(:), intent(in) :: root2
    double precision, dimension(size(theta), N+1) :: p

    integer :: K, n1
    double precision :: temp, a, b
    double precision, dimension(size(theta)) :: tc

    if (N>2700) then
      print *, 'ALFORD:big_degree'
      print *, 'Degree ', N, ' bigger than 2700'
      stop
    end if
    if (m>N) then
      print *, 'ALFORD:big_order'
      print *, 'Order ', m, ' bigger than degree ', N
      stop
    end if

    K=size(theta)
    tc=dcos(theta)
    p(1:K,1:N+1)=0d0

    p(:,m+1)=pmm(m+1,:)

    if (m .lt. N) p(:,m+2) = root2(2*m+3)*tc*p(:,m+1)

    do n1=m+3,N+1
      temp=root2(2*n1-1)/root2(n1-1+m)/root2(n1-1-m)
      a=root2(2*n1-3)*temp
      b=root2(n1+m-2)*root2(n1-m-2)*temp/root2(2*n1-5)
      p(:,n1)=a*tc*p(:,n1-1) - b*p(:,n1-2)
    end do
  end function alford

  !PRE_FFT transforms vector x of length N to vector y of length L
  ! so that:
  ! for $s=1$ we have $\sum_{m=1}^N x(m)\cos(2\pi(m-1)k)/L)
  !   = \sum_{m=1}^L y(m)\exp(-2\pi i(m-1)k/L), k=1,2,\dots,L$;
  ! for $s=-1$ we have $\sum_{m=1}^N x(m)\sin(2\pi(m-1)k/L)
  !   = i \sum_{m=1}^L y(m)\exp(-2\pi i(m-1)k/L), k=1,2,\dots,L$.
  !
  ! INPUT
  ! 1. x is a N x 1 numerical vector
  ! 2. L is a positive integer
  ! 3. s is 1 or -1. For s=1 vector  x  contains the coefficients of a cosine
  ! series. For s=-1 vector  x  contains the coefficients of a sine series.
  !
  ! OUTPUT
  ! 1. y is a L x 1 numerical vector
  ! 
  ! Notes.
  !  1. The code prepares the coefficient vector  x  for FFT

  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function pre_fft(x,L,s) result(y)
    double precision, dimension(:), intent(in) :: x
    integer, intent(in) :: L
    integer, intent(in) :: s

    double precision, dimension(L) :: y

    integer N, R, ri, riL, RL

    y=0d0

    N=size(x)
    R=ceiling(dble(N)/dble(L))-1

    do ri=0,R-1
      riL=ri*L
      y(1)=y(1)+x(riL+1)
      y(2:L)=y(2:L)+x((riL+2):(riL+L))/2d0
      y(L:2:-1)=y(L:2:-1) + dble(s)*x((riL+2):(riL+L))/2d0
    end do

    RL=R*L
    y(1)=y(1)+x(RL+1)
    y(2:N-RL)=y(2:N-RL)+x((RL+2):N)/2d0
    y(L:RL+L-N+2:-1)=y(L:RL+L-N+2:-1) + dble(s)*x((RL+2):N)/2d0

    if (s .ne. 1) y(1)=0d0

  end function pre_fft

  subroutine coeff_read(fname, cnm, snm)
    character*(*), intent(in) :: fname
    double precision, intent(out) :: cnm(:,:), snm(:,:)

    integer :: ret_i, a, b
    double precision :: c, s

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read coeff file ', fname
      stop
    end if

    do
      read(1,*,iostat=ret_i) a,b,c,s
      if(ret_i .ne. 0) exit
      cnm(a+1,b+1) = c
      snm(a+1,b+1) = s
    end do

    close(1)
  end subroutine coeff_read

  !----------------------------------------------------------------------------
  ! source for log1p from http://www.validgh.com/goldberg/addendum.html
  double precision function log1p(x)
    double precision x
    
    if (1d0 + x .eq. 1d0) then
      log1p = x
    else
      log1p = dlog(1d0 + x) * x / ((1d0 + x) - 1d0)
    end if
  end function log1p

  !----------------------------------------------------------------------------
  ! algorithm for expm1 from http://forge.scilab.org/index.php/p/distfun/
  !   source/tree/30/src/fortran/rexp.f
  !
  double precision function expm1(x)
    double precision x

    double precision p1,p2,q1,q2,q3,q4,w
    data p1/0.914041914819518d-09/, p2/0.238082361044469d-01/, &
         q1/-.499999999085958d+00/, q2/0.107141568980644d+00/, &
         q3/-.119041179760821d-01/, q4/0.595130811860248d-03/

    if (abs(x) .le. 0.15d0) then
      expm1 = x* (((p2*x+p1)*x+1d0)/ ((((q4*x+q3)*x+q2)*x+q1)*x+1d0))
    else
      w = dexp(x)
      if (x .le. 0d0) then
        expm1 = (w-0.5d0) - 0.5d0
      else
        expm1 = w * (0.5d0+ (0.5d0-1.0d0/w))
      end if
    end if
  end function expm1


      SUBROUTINE FOURT(DATT,NN,NDIM,ISIGN,IFORM,WORK,IDIM1,IDIM2)
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C                                                                      C
!C                             F O U R T                                C
!C                            ===========                               C
!C                                                                      C
!C                                                                      C
!C               THE COOLEY-TUKEY FAST FOURIER TRANSFORM                C
!C              =========================================               C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     TRANSFORM(J1,J2,,,,) = SUM(DATT(I1,I2,,,,) * W1**((I1-1)*J1-1))  C
!C                            * W2**((I2-1)*(J2-1)) * ,,,,) ,           C
!C                                                                      C
!C     WHERE I1 AND J1 RUN FROM 1 TO NN(1) AND W1=EXP(ISIGN*2*PI*       C
!C     SQRT(-1)/NN(1) ), ETC.                                           C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     THERE IS NO LIMIT IN THE DIMENSIONALITY (NUMBER OF SUBSCRIPTS)   C
!C     OF THE DATA ARRAY. IF AN INVERSE TRANSFORM (ISIGN=+1) IS         C
!C     PERFORMED UPON AN ARRAY OF TRANSFORMED DATA (ISIGN=-1), THE      C
!C     ORIGINAL DATA MULTIPLIED BY NN(1)*NN(2)*,,,, WILL REAPPEAR.      C
!C                                                                      C
!C     THE TRANSFORM VALUES ARE ALWAYS COMPLEX AND ARE RETURNED IN THE  C
!C     ORIGINAL ARRAY OF DATT, REPLACING THE INPUT DATA. THE LENGTH OF  C
!C     EACH DIMENSION OF THE DATA ARRAY MAY BE ANY INTEGER. THE PROGRAM C
!C     RUNS FASTER ON COMPOSITE INTEGERS THAN ON PRIMES, AND IS         C
!C     PARTICULARLY FAST ON NUMBERS RICH IN FACTORS OF TWO.             C
!C                                                                      C
!C     DETAILS ABOUT THE TIMING DEPENDING ON THE USED DIMENSIONS        C
!C     MAY BE FOUND IN THE OPEN FILE REPORT 83-237 FROM THOMAS G.       C
!C     HILDENBRAND (SEE BELOW).                                         C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     INPUT PARAMETER DESCRIPTION...                                   C
!C     ==============================                                   C
!C                                                                      C
!C     DATT...      IS THE ARRAY USED TO HOLD THE REAL AND IMAGINARY    C
!C                  PARTS OF THE DATA ON INPUT AND THE TRANSFORM VALUES C
!C                  ON OUTPUT. IT IS A MULTIDIMENSIONAL FLOATING POINT  C
!C                  ARRAY, WITH THE REAL AND IMAGINARY PARTS OF A DATUM C
!C                  STORED IMMEDIATELY ADJACENT IN STORAGE (SUCH AS     C
!C                  FORTRAN PLACES THEM). NORMAL FORTRTAN ORDERING IS   C
!C                  EXPECTED, THE FIRST SUBSCRIPT CHANGING FASTEST.     C
!C     NN...        THIS ARRAY OF LENGTH NDIM CONTAINS THE DIMENSIONS   C
!C                  OF THE MULTIVARIATE DATA.                           C
!C     NDIM...      NDIM SPECIFIES THE NUMBER OF DIMENSIONS IN THE      C
!C                  MULTIVARIATE DATA                                   C
!C     ISIGN...     ISIGN = -1 INDICATES A FORWARD TRANSFORM (EXPONEN-  C
!C                  TIAL SIGN IS -) AND +1 AN INVERSE TRANSFORM (SIGN   C
!C                  IS +).                                              C
!C     IFORM...     IS +1 IF THE DATA ARE COMPLEX AND 0 IF THE DATA ARE C
!C                  REAL. IF IT IS 0, THE IMAGINARY PARTS OF THE DATA   C
!C                  MUST BE SET TO ZERO.                                C
!C     WORK...      ARRAY USED FOR WORKING STORAGE. IT IS ONE-          C
!C                  DIMENSIONAL AND HAS TO BE AT LEAST OF LENGTH        C
!C                  EQUAL TO TWICE THE LARGEST ARRAY DIMENSION NN(I)    C
!C                  THAT IS NOT A POWER OF TWO. IF ALL NN(I) ARE POWERS C
!C                  OF TWO, IT IS NOT NEEDED AND MAY BE REPLACED BY ZEROC
!C                  IN THE CALLING SEQUENCE. THUS, FOR A ONE-DIMENSIONALC
!C                  ARRAY, NN(1) ODD,  WORK OCCUPIES AS MANY STORAGE    C
!C                  LOCATIONS AS DATT.                                  C
!C     IDIM1...     DIMENSION OF ARRAY DATT AS DECLARED IN THE CALLING  C
!C                  PROGRAM                                             C
!C     IDIM2...     DIMENSION OF ARRAY WORK AS DECLARED IN THE CALLING  C
!C                  PROGRAM                                             C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     EXAMPLE...                                                       C
!C     ==========                                                       C
!C                                                                      C
!C     THREE-DIMENSIONAL FORWARD FOURIER TRANSFORM OF A COMPLEX ARRAY   C
!C     DIMENSIONED 32 BY 25 BY 13 IN FORTRAN.                           C
!C                                                                      C
!C                                                                      C
!C     DIMENSION DATA(32,25,13), WORK(50), NN(3)                        C
!C     COMPLEX DATA                                                     C
!C     DATA NN/32,25,13/                                                C
!C     IDIM1=2*32*25*13                                                 C
!C     IDIM2=50                                                         C
!C     DO 1 I=1,32                                                      C
!C     DO 1 J=1,25                                                      C
!C     DO 1 K=1,13                                                      C
!C   1 DATA(I,J,K)=COMPLEX VALUE                                        C
!C     CALL FOURT(DATA,NN,3,-1,1,WORK,IDIM1,IDIM2)                      C
!C                                                                      C
!C     OR...                                                            C
!C                                                                      C
!C     DIMENSION DATA(2,32,25,13), WORK(50), NN(3)                      C
!C     DATA NN/32,25,13/                                                C
!C     IDIM1=2*32*25*13                                                 C
!C     IDIM2=50                                                         C
!C     DO 1 I=1,32                                                      C
!C     DO 1 J=1,25                                                      C
!C     DO 1 K=1,13                                                      C
!C     DATA(1,I,J,K)=REAL PART                                          C
!C   1 DATA(2,I,J,K)=IMAGINARY PART                                     C
!C     CALL FOURT(DATA,NN,3,-1,1,WORK,IDIM1,IDIM2)                      C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     THIS SUBROUTINE HAS BEEN KINDLY PROVIDED BY RENE FORSBERG FROM   C
!C     DANISH GEODETIC INSTITUTE, COPENHAGEN. THE CODE IS IN LARGE PARTSC
!C     IDENTICAL WITH A SUBROUTINE PUBLISHED BY THOMAS G. HILDENBRAND:  C
!C     "FFTFIL: A FILTERING PROGRAM BASED ON TWO-DIMENSIONAL FOURIER    C
!C     ANALYSIS OF GEOPHYSICAL DATA", UNITED STATES DEPARTMENT OF THE   C
!C     INTERIOR, US GEOLOGICAL SURVEY, OPEN FILE REPORT 83-237. THE     C
!C     MAIN DIFFERENCE OF THE TWO VERSIONS ARE OCCURING IN THE PART     C
!C     "MAIN LOOP FOR FACTORS NOT EQUAL TO TWO".                        C
!C                                                                      C
!C     THE SUBROUTINE HAS BEEN TESTED AGAINST A SIMILAR ROUTINE FROM    C
!C     THE NAG LIBRARY FOR DIMENSIONS UP TO THREE. NO DIFFERENCES WERE  C
!C     FOUND. THIS SUBROUTINE IS RUNNING FASTER THAN THE NAG-ROUTINE    C
!C     WHEN DIMENSIONS ARE POWERS OF TWO, BUT IT MAY BE A FACTOR OF     C
!C     2 ... 4 SLOWER THAN THE NAG-ROUTINE IF NO GOOD PRIME             C
!C     FACTORIZATION IS POSSIBLE.                                       C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     PROGRAM TEST PERFORMED BY    H. DENKER                27/10/1986 C
!C     =========================    INSTITUT FUER ERDMESSUNG            C
!C                                  UNIVERSITAET HANNOVER               C
!C                                  NIENBURGER STRASSE 6                C
!C                                  D-3000 HANNOVER 1                   C
!C                                                                      C
!C     LAST MODIFICATION BY         H. DENKER                26/07/2001 C
!C                                  (CONSTANTS CHANGED TO DOUBLE)       C
!C                                                                      C
!C----------------------------------------------------------------------C
!C                                                                      C
!C     COMMENT FROM RENE FORSBERG                                       C
!C        VERSION=740301                                                C
!C        PROGRAM DESCRIPTION NORSAR N-PD9 DATED 1 JULY 1970            C
!C        AUTHOR N M BRENNER                                            C
!C        FURTHER DESCRIPTION    THREE FORTRAN PROGRAMS ETC.            C
!C        ISSUED BY LINCOLN LABORATORY, MIT, JULY 1967                  C
!C        TWO CORRECTIONS BY HJORTENBERG 1974                           C
!C     THE FAST FOURIER TRANSFORM IN USASI BASIC FORTRAN                C
!C                                                                      C
!C     MODIFIED TO RC FORTRAN RF JUNE 84                                C
!C                                                                      C
!C**********************************************************************C
      implicit real*8(a-h,o-z)
      implicit integer(i-n)
      real*8 THETA
      DIMENSION DATT(IDIM1),NN(NDIM),IFACT(32),WORK(IDIM2)
!C     LEVEL 2,DATT
      NP0=0
      NPREV=0
      TWOPI=8.d0*ATAN(1.d0)
      RTHLF=1.d0/SQRT(2.d0)
      IF(NDIM-1)920,1,1
1     NTOT=2
      DO 2 IDIM=1,NDIM
      IF(NN(IDIM))920,920,2
2     NTOT=NTOT*NN(IDIM)
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     MAINLOOP FOR EACH DIMENSION                                      C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      NP1=2
      DO 910 IDIM=1,NDIM
      N=NN(IDIM)
      NP2=NP1*N
      IF(N-1)920,900,5
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     IS N A POWER OF TWO AND IF NOT, WHAT ARE ITS FACTORS             C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
5     M=N
      NTWO=NP1
      IF=1
      IDIV=2
10    IQUOT=M/IDIV
      IREM=M-IDIV*IQUOT
      IF(IQUOT-IDIV)50,11,11
11    IF(IREM)20,12,20
12    NTWO=NTWO+NTWO
      IFACT(IF)=IDIV
      IF=IF+1
      M=IQUOT
      GO TO 10
20    IDIV=3
      INON2=IF
30    IQUOT=M/IDIV
      IREM=M-IDIV*IQUOT
      IF(IQUOT-IDIV)60,31,31
31    IF(IREM)40,32,40
32    IFACT(IF)=IDIV
      IF=IF+1
      M=IQUOT
      GO TO 30
40    IDIV=IDIV+2
      GO TO 30
50    INON2=IF
      IF(IREM)60,51,60
51    NTWO=NTWO+NTWO
      GO TO 70
60    IFACT(IF)=M
70    NON2P=NP2/NTWO
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     SEPARATE FOUR CASES---                                           C
!C        1. COMPLEX TRANSFORM                                          C
!C        2. REAL TRANSFORM FOR THE 2ND, 3ND, ETC. DIMENSION.  METHOD-- C
!C           TRANSFORM HALF THE DATT, SUPPLYING THE OTHER HALF BY CON-  C
!C           JUGATE SYMMETRY.                                           C
!C        3. REAL TRANSFORM FOR THE 1ST DIMENSION,N ODD.  METHOD--      C
!C           SET THE IMAGINARY PARTS TO ZERO                            C
!C        4. REAL TRANSFORM FOR THE 1ST DIMENSION,N EVEN.METHOD--       C
!C           TRANSFORM A COMPLEX ARRAY OF LENGHT N/2 WHOSE REAL PARTS   C
!C           ARE THE EVEN NUMBERD REAL VALUES AND WHOSE IMAGINARY PARTS C
!C           ARE THE ODD NUMBEREDREAL VALUES.  SEPARATE AND SUPPLY      C
!C           THE SECOND HALF BY CONJUGATE SUMMETRY.                     C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      ICASE=1
      IFMIN=1
      I1RNG=NP1
      IF(IDIM-4)74,100,100
74    IF(IFORM)71,71,100
71    ICASE=2
      I1RNG=NP0*(1+NPREV/2)
      IF(IDIM-1)72,72,100
72    ICASE=3
      I1RNG=NP1
      IF(NTWO-NP1)100,100,73
73    ICASE=4
      IFMIN=2
      NTWO=NTWO/2
      N=N/2
      NP2=NP2/2
      NTOT=NTOT/2
      I=1
      DO 80 J=1,NTOT
      DATT(J)=DATT(I)
80    I=I+2
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     SHUFFLE DATT BY BIT REVERSAL, SINCE N=2**K.  AS THE SHUFFLING    C
!C     CAN BE DONE BY SIMPLE INTERCHANGE, NO WORKING ARRAY IS NEEDED    C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
100   IF(NON2P-1)101,101,200
101   NP2HF=NP2/2
      J=1
      DO 150 I2=1,NP2,NP1
      IF(J-I2)121,130,130
121   I1MAX=I2+NP1-2
      DO 125 I1=I2,I1MAX,2
      DO 125 I3=I1,NTOT,NP2
      J3=J+I3-I2
      TEMPR=DATT(I3)
      TEMPI=DATT(I3+1)
      DATT(I3)=DATT(J3)
      DATT(I3+1)=DATT(J3+1)
      DATT(J3)=TEMPR
125   DATT(J3+1)=TEMPI
130   M=NP2HF
140   IF(J-M)150,150,141
141   J=J-M
      M=M/2
      IF(M-NP1)150,140,140
150   J=J+M
      GO TO 300
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     SHUFFLE DATT BY DIGIT REVERSAL FOR GENERAL N                     C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
200   NWORK=2*N
      DO 270 I1=1,NP1,2
      DO 270 I3=I1,NTOT,NP2
      J=I3
      DO 260 I=1,NWORK,2
      IF(ICASE-3)210,220,210
210   WORK(I)=DATT(J)
      WORK(I+1)=DATT(J+1)
      GO TO 240
220   WORK(I)=DATT(J)
      WORK(I+1)=0.d0
240   IFP2=NP2
      IF=IFMIN
250   IFP1=IFP2/IFACT(IF)
      J=J+IFP1
      IF(J-I3-IFP2)260,255,255
255   J=J-IFP2
      IFP2=IFP1
      IF=IF+1
      IF(IFP2-NP1)260,260,250
260   CONTINUE
      I2MAX=I3+NP2-NP1
      I=1
      DO 270 I2=I3,I2MAX,NP1
      DATT(I2)=WORK(I)
      DATT(I2+1)=WORK(I+1)
270   I=I+2
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     MAIN LOOP FOR FACTORS OF TWO                                     C
!C     W=EXP(ISIGN*2*PI*SQRT(-1)*M/(4*MMAX)). CHECK FOR W=ISIGN*SQRT(-1)C
!C     AND REPEAT FOR W=W*(1+ISIGN*SQRT(-1))/SQRT(2)                    C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
300   IF(NTWO-NP1)600,600,305
305   NP1TW=NP1+NP1
      IPAR=NTWO/NP1
310   IF(IPAR-2)350,330,320
320   IPAR=IPAR/4
      GO TO 310
330   DO 340 I1=1,I1RNG,2
      DO 340 K1=I1,NTOT,NP1TW
      K2=K1+NP1
      TEMPR=DATT(K2)
      TEMPI=DATT(K2+1)
      DATT(K2)=DATT(K1)-TEMPR
      DATT(K2+1)=DATT(K1+1)-TEMPI
      DATT(K1)=DATT(K1)+TEMPR
340   DATT(K1+1)=DATT(K1+1)+TEMPI
350   MMAX=NP1
360   IF(MMAX-NTWO/2)370,600,600
370   LMAX=MAX(NP1TW,MMAX/2)
      DO 570 L=NP1,LMAX,NP1TW
      M=L
      IF(MMAX-NP1)420,420,380
380   THETA=-TWOPI*L/(4.d0*MMAX)
      IF(ISIGN)400,390,390
390   THETA=-THETA
400   WR=COS(THETA)
      WI=SIN(THETA)
410   W2R=WR*WR-WI*WI
      W2I=2.d0*WR*WI
      W3R=W2R*WR-W2I*WI
      W3I=W2R*WI+W2I*WR
420   DO 530 I1=1,I1RNG,2
      KMIN=I1+IPAR*M
      IF(MMAX-NP1)430,430,440
430   KMIN=I1
440   KDIF=IPAR*MMAX
450   KSTEP=4*KDIF
      IF(KSTEP-NTWO)460,460,530
460   DO 520 K1=KMIN,NTOT,KSTEP
      K2=K1+KDIF
      K3=K2+KDIF
      K4=K3+KDIF
      IF(MMAX-NP1)470,470,480
470   U1R=DATT(K1)+DATT(K2)
      U1I=DATT(K1+1)+DATT(K2+1)
      U2R=DATT(K3)+DATT(K4)
      U2I=DATT(K3+1)+DATT(K4+1)
      U3R=DATT(K1)-DATT(K2)
      U3I=DATT(K1+1)-DATT(K2+1)
      IF(ISIGN)471,472,472
471   U4R=DATT(K3+1)-DATT(K4+1)
      U4I=DATT(K4)-DATT(K3)
      GO TO 510
472   U4R=DATT(K4+1)-DATT(K3+1)
      U4I=DATT(K3)-DATT(K4)
      GO TO 510
480   T2R=W2R*DATT(K2)-W2I*DATT(K2+1)
      T2I=W2R*DATT(K2+1)+W2I*DATT(K2)
      T3R=WR*DATT(K3)-WI*DATT(K3+1)
      T3I=WR*DATT(K3+1)+WI*DATT(K3)
      T4R=W3R*DATT(K4)-W3I*DATT(K4+1)
      T4I=W3R*DATT(K4+1)+W3I*DATT(K4)
      U1R=DATT(K1)+T2R
      U1I=DATT(K1+1)+T2I
      U2R=T3R+T4R
      U2I=T3I+T4I
      U3R=DATT(K1)-T2R
      U3I=DATT(K1+1)-T2I
      IF(ISIGN)490,500,500
490   U4R=T3I-T4I
      U4I=T4R-T3R
      GO TO 510
500   U4R=T4I-T3I
      U4I=T3R-T4R
510   DATT(K1)=U1R+U2R
      DATT(K1+1)=U1I+U2I
      DATT(K2)=U3R+U4R
      DATT(K2+1)=U3I+U4I
      DATT(K3)=U1R-U2R
      DATT(K3+1)=U1I-U2I
      DATT(K4)=U3R-U4R
520   DATT(K4+1)=U3I-U4I
      KDIF=KSTEP
      KMIN=4*(KMIN-I1)+I1
      GO TO 450
530   CONTINUE
      M=M+LMAX
      IF(M-MMAX)540,540,570
540   IF(ISIGN)550,560,560
550   TEMPR=WR
      WR=(WR+WI)*RTHLF
      WI=(WI-TEMPR)*RTHLF
      GO TO 410
560   TEMPR=WR
      WR=(WR-WI)*RTHLF
      WI=(TEMPR+WI)*RTHLF
      GO TO 410
570   CONTINUE
      IPAR=3-IPAR
      MMAX=MMAX+MMAX
      GO TO 360
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     MAIN LOOP FOR FACTOERS NOT EQUAL TO TWO                          C
!C     W=EXP(ISIGN*2*PI*SQRT(-1)*(J1+J2-I3-1)/IFP2)                     C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
600   IF(NON2P-1)700,700,601
601   IFP1=NTWO
      IF=INON2
610   IFP2=IFACT(IF)*IFP1
      THETA=-TWOPI/IFACT(IF)
      IF(ISIGN)612,611,611
611   THETA=-THETA
612   WSTPR=COS(THETA)
      WSTPI=SIN(THETA)
      DO 650 J1=1,IFP1,NP1
      THETM=-TWOPI*(J1-1.d0)/IFP2
      IF(ISIGN)614,613,613
613   THETM=-THETM
614   WMINR=COS(THETM)
      WMINI=SIN(THETM)
      I1MAX=J1+I1RNG-2
      DO 650 I1=J1,I1MAX,2
      DO 650 I3=I1,NTOT,NP2
      I=1
      WR=WMINR
      WI=WMINI
      J2MAX=I3+IFP2-IFP1
      DO 640 J2=I3,J2MAX,IFP1
      TWOWR=WR+WR
      J3MAX=J2+NP2-IFP2
      DO 630 J3=J2,J3MAX,IFP2
      JMIN=J3-J2+I3
      J=JMIN+IFP2-IFP1
      SR=DATT(J)
      SI=DATT(J+1)
      OLDSR=0.d0
      OLDSI=0.d0
      J=J-IFP1
620   STMPR=SR
      STMPI=SI
      SR=TWOWR*SR-OLDSR+DATT(J)
      SI=TWOWR*SI-OLDSI+DATT(J+1)
      OLDSR=STMPR
      OLDSI=STMPI
      J=J-IFP1
      IF(J-JMIN)621,621,620
621   WORK(I)=WR*SR-WI*SI-OLDSR+DATT(J)
      WORK(I+1)=WI*SR+WR*SI-OLDSI+DATT(J+1)
630   I=I+2
      WTEMP=WR*WSTPI
      WR=WR*WSTPR-WI*WSTPI
640   WI=WI*WSTPR+WTEMP
      I=1
      DO 650 J2=I3,J2MAX,IFP1
      J3MAX=J2+NP2-IFP2
      DO 650 J3=J2,J3MAX,IFP2
      DATT(J3)=WORK(I)
      DATT(J3+1)=WORK(I+1)
650   I=I+2
      IF=IF+1
      IFP1=IFP2
      IF(IFP1-NP2)610,700,700
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     COMPLETE AREAL TRANSFORM IN THE 1ST DIMENSION, N EVEN, BY CON-   C
!C     JUGATE SYMMETRIES                                                C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
700   GO TO (900,800,900,701),ICASE
701   NHALF=N
      N=N+N
      THETA=-TWOPI/N
      IF(ISIGN)703,702,702
702   THETA=-THETA
703   WSTPR=COS(THETA)
      WSTPI=SIN(THETA)
      WR=WSTPR
      WI=WSTPI
      IMIN=3
      JMIN=2*NHALF-1
      GO TO 725
710   J=JMIN
      DO 720 I=IMIN,NTOT,NP2
      SUMR=(DATT(I)+DATT(J))/2.d0
      SUMI=(DATT(I+1)+DATT(J+1))/2.d0
      DIFR=(DATT(I)-DATT(J))/2.d0
      DIFI=(DATT(I+1)-DATT(J+1))/2.d0
      TEMPR=WR*SUMI+WI*DIFR
      TEMPI=WI*SUMI-WR*DIFR
      DATT(I)=SUMR+TEMPR
      DATT(I+1)=DIFI+TEMPI
      DATT(J)=SUMR-TEMPR
      DATT(J+1)=-DIFI+TEMPI
720   J=J+NP2
      IMIN=IMIN+2
      JMIN=JMIN-2
      WTEMP=WR*WSTPI
      WR=WR*WSTPR-WI*WSTPI
      WI=WI*WSTPR+WTEMP
725   IF(IMIN-JMIN)710,730,740
730   IF(ISIGN)731,740,740
731   DO 735 I=IMIN,NTOT,NP2
735   DATT(I+1)=-DATT(I+1)
740   NP2=NP2+NP2
      NTOT=NTOT+NTOT
      J=NTOT+1
      IMAX=NTOT/2+1
745   IMIN=IMAX-2*NHALF
      I=IMIN
      GO TO 755
750   DATT(J)=DATT(I)
      DATT(J+1)=-DATT(I+1)
755   I=I+2
      J=J-2
      IF(I-IMAX)750,760,760
760   DATT(J)=DATT(IMIN)-DATT(IMIN+1)
      DATT(J+1)=0.d0
      IF(I-J)770,780,780
765   DATT(J)=DATT(I)
      DATT(J+1)=DATT(I+1)
770   I=I-2
      J=J-2
      IF(I-IMIN)775,775,765
775   DATT(J)=DATT(IMIN)+DATT(IMIN+1)
      DATT(J+1)=0.d0
      IMAX=IMIN
      GO TO 745
780   DATT(1)=DATT(1)+DATT(2)
      DATT(2)=0.d0
      GO TO 900
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     COMPLETE A REAL TRANSFORM FOR THE 2ND, 3RD, ETC. DIMENSION BY    C
!C     CONJUGATE SYMMETRIES.                                            C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
800   IF(I1RNG-NP1)805,900,900
805   DO 860 I3=1,NTOT,NP2
      I2MAX=I3+NP2-NP1
      DO 860 I2=I3,I2MAX,NP1
      IMAX=I2+NP1-2
      IMIN=I2+I1RNG
      JMAX=2*I3+NP1-IMIN
      IF(I2-I3)820,820,810
810   JMAX=JMAX+NP2
820   IF(IDIM-2)850,850,830
830   J=JMAX+NP0
      DO 840 I=IMIN,IMAX,2
      DATT(I)=DATT(J)
      DATT(I+1)=-DATT(J+1)
840   J=J-2
850   J=JMAX
      DO 860 I=IMIN,IMAX,NP0
      DATT(I)=DATT(J)
      DATT(I+1)=-DATT(J+1)
860   J=J-NP0
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C     END OF LOOP ON EACH DIMENSION                                    C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
900   NP0=NP1
      NP1=NP2
910   NPREV=N
920   RETURN
      end subroutine fourt

end program emmsynth_init
