program emmsynth_standard
  implicit none
  !EMMSYNTH_STANDARD computes and writes in a file
  ! the values of the magnetic field gradient at scattered points
  ! by the standard algorithm based on the EMM2015 or EMM2017 coefficients
  ! and compares these values with the output of EMMSYNTH_FAST:
  ! given in 'scattered_points_values.dat'.
  !
  ! INPUT:
  ! 1. free format files
  !    'EMM2015.COF' and 'EMM2015SV.COF' for Enhanced Magnetic Model EMM2015 or
  !    'EMM2017.COF' and 'EMM2017SV.COF' for Enhanced Magnetic Model EMM2017.
  !    with the model coefficients
  ! 2. free format file 'scattered_points_values.dat' with point's geodetic
  !    latitude, longitude, (in degrees), height above the reference
  !    ellipsoid (in meters) and the value of the 10 magnetic field components
  !    computed by EMMSYNTH_FAST.
  ! 3. The chosen magnetic model.
  ! 4. The parameter 'year' when the model is computed.
  !
  ! OUTPUT:
  ! 1. free format file 'scattered_points_values_standard.dat' with point's
  !    geodetic latitude, longitude, (in degrees), height (in meters)
  !    and the values of the 10 magnetic field components evaluated by
  !    EMMSYNTH_STANDARD.
  ! 2. On screen statistics, including the 'Maximal absolute errors'.
  !
  ! Notes.
  ! 1. The code applies different algorithms compared to the standard EMM
  !    code in the computation of 'Xprime' and 'Yprime'.
  !    These algorithms are stable all over the sphere including the poles.
  ! 2. The fields in every row of 'scattered_points_values.dat' and
  !    'scattered_points_values_standard.dat' contain the values of the
  !    the magnetic field gradient elements given in  'component'. 'phi' and
  !    'lambda' are geodetic geographic latitude and longitude (in degrees),
  !    'height' is the height above the reference ellipsoid (in meters).
  !
  ! Version: November 2018
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  integer :: model, modeldegree
  character(len=80) :: s1, s1SV, s2, s3
  double precision ::year, year0, year1, dt
  integer :: icount, icount_rate, ret_i
  real*8 :: time0, time1, time2, time3, time4, time5, time6
  real*8 :: TTime0, TTime1, TTime2, TTime3, TTime4, TTime5
  character(len=24), dimension(13) :: component
  double precision :: ae, inv_flat, be, am
  integer :: npt
  double precision, dimension(:,:), allocatable :: pt, values, point, difference
  double precision, dimension(:), allocatable :: phi, phiprime
  integer :: nmax, nmax1
  double precision, dimension(:,:), allocatable :: cnm, snm
  double precision, dimension(:,:), allocatable :: cnmw, snmw
  double precision, dimension(:,:), allocatable :: cnm1, snm1
  double precision, dimension(:,:), allocatable :: cnm2, snm2
  double precision, dimension(:,:), allocatable :: cnm3, snm3
  double precision, dimension(:,:), allocatable :: cnm4, snm4
  double precision, dimension(:,:), allocatable :: cnm5, snm5
  integer :: NSV, NSV1
  integer :: n, mm, i, m
  double precision :: tmp, tmp1
  double precision, dimension(:), allocatable :: w, w1, w2
  double precision, dimension(:), allocatable :: dphi, cdphi, sdphi
  double precision, dimension(:), allocatable :: Horiz2
  integer :: nerr
  double precision :: pi
  character(len=80) :: debugdir

  parameter (pi = 3.14159265358979323846264338327950288d0)

  ! model and year input and check
  write(*, fmt='(A)', advance="no") &
    'Which model to use? (Enter 1 for EMM2015 or 2 for EMM2017): '
  read (*,*) model

  select case(model)
    case(1)
      modeldegree=740
      s1='EMM2015.COF'
      s1SV='EMM2015SV.COF'
      year0=2015.0
      year1=2020.0

    case(2)
      modeldegree=790
      s1='EMM2017.COF'
      s1SV='EMM2017SV.COF'
      year0=2017.0
      year1=2022.0

    case default
      print *, 'emmsynth_standard:model'
      print *, 'wrong input"'
      stop
  end select

  write(*, fmt='(A)', advance="no") &
    'Enter decimal year for geomagnetic computations: '
  read (*,*) year

  if((year.lt.year0).or.(year.gt.year1)) then
    print *, 'emmsynth_standard:year'
    print *, 'The year must be decimal number between ', year0,' and ', year1
    stop
  end if
  dt=year-year0

  call system_clock(icount, icount_rate)
  time0=dble(icount)/dble(icount_rate)

  component= (/ &
    'phi                     ', &
    'lambda                  ', &
    'height                  ', &
    'Xprime:                 ', &
    'Yprime:                 ', &
    'Zprime:                 ', &
    'North component X:      ', &
    'East component Y:       ', &
    'Down component Z:       ', &
    'Horizontal intensity H: ', &
    'Total intensity F:      ', &
    'Inclination I:          ', &
    'Declination D           ' /)


  ! Earth constants
  ! wgs84_constants
  ae = 6378137d0
  inv_flat=298.257223563d0
  be = ae*(1d0-1d0/inv_flat)

  am=6371200d0 !m; geomagnetic reference radius

  ! scattered points for solid spherical polynomial evaluation
  call system_clock(icount)
  time1=dble(icount)/dble(icount_rate)

  s2='scattered_points_values.dat' !scattered points input file name
  npt=dlmsize(s2)
  allocate(pt(npt,13), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: pt'
    stop
  end if

  call dlmread13(s2, npt, pt)

  ! checking point geodetic coordinates
  if (any(pt(1:npt,1).lt.-90d0) .or. any(pt(1:npt,1).gt.90d0)) then
    print *, 'emmsynth_standard:pt'
    print *, 'latitude must be between -90 and 90 degrees'
    stop
  end if
  if (any(pt(1:npt,2).lt.-180d0) .or. any(pt(1:npt,2).gt.180d0)) then
    print *, 'emmsynth_standard:pt'
    print *, 'longitude must be between -180 and 180 degrees'
    stop
  end if
  if (any(pt(1:npt,3).lt.-415d0) .or. any(pt(1:npt,3).gt.1d6)) then
    print *, 'emmsynth_standard:pt'
    print *, 'hight must be between -415 m and 1,000,000 m'
    stop
  end if

  ! output array
  allocate(values(npt,13), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: values'
    stop
  end if
  values = 0d0
  values(1:npt, 1:3) = pt(1:npt, 1:3)

  ! work array
  allocate(point(npt,13), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: point'
    stop
  end if

  point=0d0
  point(1:npt,2)=pt(1:npt,2)*pi/180d0
  where (point(1:npt, 2).lt.0d0) point(1:npt,2)=point(1:npt,2)+2d0*pi

  allocate(phi(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: phi'
    stop
  end if
  allocate(phiprime(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: phiprime'
    stop
  end if
  phi=pt(1:npt,1)*pi/180d0

  call gd2gc(phi, pt(1:npt,3), ae, be, &
    phiprime, point(1:npt,3))

  point(1:npt,1) = pi/2d0 - phiprime !co-latitude

  point(1:npt,3) = am/point(1:npt,3)

  call system_clock(icount)
  time2=dble(icount)/dble(icount_rate)
  TTime1=time2-time1 !coordinate transformation time

  ! Read spherical harmonic geomagnetic model

  nmax=modeldegree
  nmax1=nmax+1

  allocate(cnm(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnm'
    stop
  end if
  allocate(snm(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snm'
    stop
  end if

  cnm(1:nmax1,1:nmax1) = 0d0
  snm(1:nmax1,1:nmax1) = 0d0

  call coeff_read(s1, cnm, snm)

  NSV=15
  NSV1=NSV+1

  allocate(cnmw(NSV1,NSV1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnmw'
    stop
  end if
  allocate(snmw(NSV1,NSV1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snmw'
    stop
  end if

  cnmw(1:NSV1,1:NSV1) = 0d0
  snmw(1:NSV1,1:NSV1) = 0d0

  call coeff_read(s1SV, cnmw, snmw)

  cnm(1:NSV1,1:NSV1)=cnm(1:NSV1,1:NSV1)+dt*cnmw(1:NSV1,1:NSV1)
  snm(1:NSV1,1:NSV1)=snm(1:NSV1,1:NSV1)+dt*snmw(1:NSV1,1:NSV1)

  deallocate(snmw)
  deallocate(cnmw)

  call system_clock(icount)
  time3=dble(icount)/dble(icount_rate)
  TTime2=time3-time2 !model read time

  ! from Schmidt semi-normalized ALF to fully-normalized ALF
  do n=1,nmax1
    tmp = dble(2*(n-1)+1)**(-0.5d0)
    cnm(n,1:nmax1) = cnm(n,1:nmax1)*tmp
    snm(n,1:nmax1) = snm(n,1:nmax1)*tmp
  end do

  ! Spherical coefficient formation

  ! Xprime

  allocate(cnm1(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnm1'
    stop
  end if
  allocate(snm1(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snm1'
    stop
  end if
  allocate(cnm2(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnm2'
    stop
  end if
  allocate(snm2(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snm2'
    stop
  end if

  cnm1(1:nmax1,1:nmax1) = 0d0
  snm1(1:nmax1,1:nmax1) = 0d0
  cnm2(1:nmax1,1:nmax1) = 0d0
  snm2(1:nmax1,1:nmax1) = 0d0

  tmp=1d0
  allocate(w(nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: w'
    stop
  end if

  do n=2,2
    w = (/ (dsqrt(dble((n-mm)*(n+mm-1)))/2d0, mm=1,n) /)
    w(1)=w(1)*dsqrt(2d0)
    cnm1(n,1)=w(1)*cnm(n,2)*tmp
    cnm1(n,n)=-w(n-1)*cnm(n,n-1)*tmp
    snm1(n,1)=0
    snm1(n,n)=0
    cnm2(n,1)=w(1)*snm(n,2)*tmp
    cnm2(n,n)=0
    snm2(n,1)=0
    snm2(n,n)=-w(n-1)*cnm(n,n-1)*tmp
  end do
  do n=3,nmax1
    snm(n,1)=0 !just in case this is not fulfilled
    w = (/ (dsqrt(dble((n-mm)*(n+mm-1)))/2d0, mm=1,n) /)
    w(1)=w(1)*dsqrt(2d0)
    cnm1(n,1)=w(1)*cnm(n,2)*tmp
    cnm1(n,2:n-1)=(w(2:n-1)*cnm(n,3:n)-w(1:n-2)*cnm(n,1:n-2))*tmp
    cnm1(n,n)=-w(n-1)*cnm(n,n-1)*tmp
    snm1(n,2:n-1)=(w(2:n-1)*snm(n,3:n)-w(1:n-2)*snm(n,1:n-2))*tmp
    snm1(n,n)=-w(n-1)*snm(n,n-1)*tmp
    cnm2(n,1)=w(1)*snm(n,2)*tmp
    cnm2(n,2:n-1)=(w(2:n-1)*snm(n,3:n)+w(1:n-2)*snm(n,1:n-2))*tmp
    cnm2(n,n)=w(n-1)*snm(n,n-1)*tmp
    snm2(n,2:n-1)=(-w(2:n-1)*cnm(n,3:n)-w(1:n-2)*cnm(n,1:n-2))*tmp
    snm2(n,n)=-w(n-1)*cnm(n,n-1)*tmp
  end do

  ! Yprime
  allocate(cnm3(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnm3'
    stop
  end if
  allocate(snm3(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snm3'
    stop
  end if
  allocate(cnm4(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnm4'
    stop
  end if
  allocate(snm4(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snm4'
    stop
  end if

  cnm3(1:nmax1,1:nmax1) = 0d0
  snm3(1:nmax1,1:nmax1) = 0d0
  cnm4(1:nmax1,1:nmax1) = 0d0
  snm4(1:nmax1,1:nmax1) = 0d0
  tmp=.5d0
  w=(/ (dsqrt(dble(i-1)*i), i=1,2*nmax1) /)
  cnm3(1:nmax,1)=-dsqrt(2d0)*w(2:nmax1)*snm(2:nmax1,2)
  cnm4(1:nmax,1)=dsqrt(2d0)*w(2:nmax1)*cnm(2:nmax1,2)
  do m=2,nmax
    cnm3(m:nmax,m)=-w(m+m:m+nmax)*snm(m+1:nmax1,m+1)
    snm3(m:nmax,m)=w(m+m:m+nmax)*cnm(m+1:nmax1,m+1)
    cnm4(m:nmax,m)=w(m+m:m+nmax)*cnm(m+1:nmax1,m+1)
    snm4(m:nmax,m)=w(m+m:m+nmax)*snm(m+1:nmax1,m+1)
  end do
  do m=3,nmax
    cnm3(m:nmax,m)=cnm3(m:nmax,m)-w(2:nmax-m+2)*snm(m+1:nmax1,m-1)
    snm3(m:nmax,m)=snm3(m:nmax,m)+w(2:nmax-m+2)*cnm(m+1:nmax1,m-1)
    cnm4(m:nmax,m)=cnm4(m:nmax,m)-w(2:nmax-m+2)*cnm(m+1:nmax1,m-1)
    snm4(m:nmax,m)=snm4(m:nmax,m)-w(2:nmax-m+2)*snm(m+1:nmax1,m-1)
  end do
  do m=1,nmax1
    tmp1 = dsqrt(dble(2*m+1)/dble(2*m-1))*tmp
    cnm3(m,:)=cnm3(m,:)*tmp1
    cnm4(m,:)=cnm4(m,:)*tmp1
    snm3(m,:)=snm3(m,:)*tmp1
    snm4(m,:)=snm4(m,:)*tmp1
  end do

  deallocate(w)

  ! Zprime
  allocate(cnm5(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cnm5'
    stop
  end if
  allocate(snm5(nmax1,nmax1), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: snm5'
    stop
  end if
  do m=1,nmax1
    cnm5(m,:)=-cnm(m,:)*m
    snm5(m,:)=-snm(m,:)*m
  end do

  deallocate(snm)
  deallocate(cnm)

  call system_clock(icount)
  time4=dble(icount)/dble(icount_rate)
  TTime3=time4-time3 !coefficient transformation time

  ! computing Xprime, Yprime, Zprime
  allocate(w1(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: w1'
    stop
  end if
  allocate(w2(npt), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: w2'
    stop
  end if

  w1=solid_polynomial_values_scattered(point(:,1),point(:,2),point(:,3), &
    cnm1,snm1,1d260)
  w2=solid_polynomial_values_scattered(point(:,1),point(:,2),point(:,3), &
    cnm2,snm2,1d260)
  values(:,4)=(w1*dcos(point(:,2))+w2*dsin(point(:,2)))*point(:,3)**2

  w1=solid_polynomial_values_scattered(point(:,1),point(:,2),point(:,3), &
    cnm3,snm3,1d260)
  w2=solid_polynomial_values_scattered(point(:,1),point(:,2),point(:,3), &
    cnm4,snm4,1d260)
  values(:,5)=(w1*dcos(point(:,2))+w2*dsin(point(:,2)))*point(:,3)**3

  values(:,6)= &
    solid_polynomial_values_scattered(point(:,1),point(:,2),point(:,3), &
      cnm5,snm5,1d260)*point(:,3)**2

  deallocate(w2)
  deallocate(w1)
  deallocate(snm5)
  deallocate(cnm5)
  deallocate(snm4)
  deallocate(cnm4)
  deallocate(snm3)
  deallocate(cnm3)
  deallocate(snm2)
  deallocate(cnm2)
  deallocate(snm1)
  deallocate(cnm1)
  deallocate(point)

  ! computing quantities in the ellipsoidal reference frame

  allocate(dphi(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: dphi'
    stop
  end if

  allocate(cdphi(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: cdphi'
    stop
  end if

  allocate(sdphi(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: sdphi'
    stop
  end if

  ! rotating geocentric to ellipsoidal reference frame
  dphi=(phiprime-phi)
  cdphi=cos(dphi)
  sdphi=sin(dphi)
  values(1:npt,7)=values(1:npt,4)*cdphi-values(1:npt,6)*sdphi
  values(1:npt,8)=values(1:npt,5)
  values(1:npt,9)=values(1:npt,4)*sdphi+values(1:npt,6)*cdphi

  deallocate(sdphi)
  deallocate(cdphi)
  deallocate(dphi)

  deallocate(phiprime)
  deallocate(phi)

  ! evaluating the remaining components of the field
  allocate(Horiz2(npt), stat = ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: Horiz2'
    stop
  end if
  Horiz2=values(:,7)**2+values(:,8)**2
  values(:,10)=dsqrt(Horiz2)
  values(:,11)=dsqrt(Horiz2+values(:,9)**2)
  values(:,12)=datan2(values(:,9),values(:,10))
  values(:,13)=datan2(values(:,8),values(:,7))

  deallocate(Horiz2)

  call system_clock(icount)
  time5=dble(icount)/dble(icount_rate)
  TTime4=time5-time4 !derived computation time

  ! output

  !scattered points and values output file name
  s3='scattered_points_values_standard.dat'

  open(unit=1, file=s3, status='replace', iostat=ret_i)
  if(ret_i .ne. 0) then
    print *, 'Cannot write output file ', s3
    stop
  end if

  do n = 1,npt
    write(1, '(2(F20.14),F16.8,8(F12.4),2(F12.8))') ( values(n,i), i=1,13 )
  end do

  close(1)

  call system_clock(icount)
  time6=dble(icount)/dble(icount_rate)
  TTime5=time6-time5

  TTime0=time6-time0 !Total time

  ! error calculation
  allocate(difference(npt,3), stat=ret_i)
  if(ret_i .gt. 0) then
    print *, 'Out of memory: difference'
    stop
  end if
  difference=values-pt

  !modify Inclination error
  nerr = count(difference(1:npt, 12).gt.0.99d0*pi) + &
    count(difference(1:npt, 12).lt.-0.99d0*pi)

  where (difference(1:npt, 12).gt.0.99d0*pi) &
    difference(1:npt,12)=difference(1:npt,12)-pi
  where (difference(1:npt, 12).lt.-0.99d0*pi) &
    difference(1:npt,12)=difference(1:npt,12)+pi

  if (nerr.gt.0) then
    print *, 'Number of points with modify Inclination error: ', nerr
  end if

  !modify Declination error
  nerr=count((difference(:,13).gt.0.99*pi).and.(difference(:,13).lt.1.01*pi))+&
    count((difference(:,13).lt.-0.99*pi).and.(difference(:,13).gt.-1.01*pi))+&
    count(difference(:,13).gt.1.98*pi) + &
    count(difference(:,13).lt.-1.98*pi)

  where((difference(:,13).gt.0.99*pi).and.(difference(:,13).lt.1.01*pi)) &
    difference(:,13)=difference(:,13)-pi
  where((difference(:,13).lt.-0.99*pi).and.(difference(:,13).gt.-1.01*pi)) &
    difference(:,13)=difference(:,13)+pi
  where(difference(:,13).gt.1.98*pi) &
    difference(:,13)=difference(:,13)-2d0*pi
  where(difference(:,13).lt.-1.98*pi) &
    difference(:,13)=difference(:,13)+2d0*pi

  if (nerr.gt.0) then
    print *,'Number of points with modify Declination error: ', nerr
  end if

  ! statistics

  write (*,'(A)') &
    '---------------------------------------------------------------'
  select case(model)
    case(1)
      write(*,'(A)') '        Program:  emmsynth_standard,       Model: EMM2015'
    case(2)
      write(*,'(A)') '        Program:  emmsynth_standard,       Model: EMM2017'
  end select

  write (*,'(A,F10.5)') &
    'Evaluation year:            = ', year
  write (*,'(A)') &
    '---------------------------------------------------------------'
  write (*,'(A,F15.5,A)') &
    'Total time                  = ', TTime0, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    'Coordinate load time        = ', TTime1, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    'Model load time             = ', TTime2, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    'Coefficient formation time  = ', TTime3, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    'Synthesis time              = ', TTime4, ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    'Values write time           = ', TTime5, ' CPU seconds'
  write (*,'(A,I9)') &
    'Total number of points      = ', npt
  write (*,'(A,F15.5,A)') &
    'Execution time per point    = ', (TTime4/real(npt)), ' CPU seconds'
  write (*,'(A,F15.5,A)') &
    'Retrieved points per second = ', (real(npt)/TTime4)
  write (*,'(A)') &
    '---------------------------------------------------------------'
  write (*,'(A)') &
    'Maximal absolute errors at the given scattered points'
  write (*,'(A)') &
    '---------------------------------------------------------------'
  do m=4,13
    write(*, '(A,F15.5)') component(m), maxval(dabs(difference(:,m)))
  end do
  write (*,'(A,F15.5,A)') &
    '---------------------------------------------------------------'

  deallocate(difference)
  deallocate(values)
  deallocate(pt)

  contains

  ! SOLID_POLYNOMIAL_VALUES_SCATTERED computes the values 
  ! of a solid spherical polynomial with coefficients in  cnm, snm
  ! at the points (theta,lambda,ar)
  !
  ! INPUT
  ! 1. theta, lambda, ar  are arrays of equal length with the spherical
  !   coordinates of the scattered points in the unit 3-d ball.
  !   theta is co-latitude, lambda  is longitude,  
  !   ar  is distance from the origin.
  ! 2. cnm, snm  are lower square matrixes of the same size
  !   with the cosine and sine coefficients of the solid polynomial
  ! 3. gs is a global scaling factor for avoiding underflow in evaluating the
  !   associated Legendre functions, say gs=10^260.
  !
  ! OUTPUT
  ! 1. y  is a vector of the lenth of  theta  with the polynomial values
  !
  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  function solid_polynomial_values_scattered(theta,lambda,ar,cnm,snm,gs) &
      result(y)
    double precision, dimension(:), intent(in) :: theta, lambda, ar
    double precision, dimension(:,:), intent(in) :: cnm, snm
    double precision, intent(in) :: gs

    double precision, dimension(size(theta)) :: y

    integer :: N1, nmax, KK1, k, m, i, j
    double precision :: lm
    double precision, dimension(:), allocatable :: root2
    double precision, dimension(size(theta), size(cnm,1)) :: kappa, Cmt, Smt
    double precision, dimension(size(theta), size(cnm,1)) :: p, pk
    double precision, dimension(size(cnm,1), size(theta)) :: pmm

    N1=size(cnm,1)
    nmax=N1-1

    allocate(root2(2*nmax+5), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: root2'
      stop
    end if
    root2=(/ (dsqrt(dble(i)), i=1,2*nmax+5) /)

    KK1=size(theta)

    kappa(:,1) = 1d0
    do k=1,nmax
      kappa(:,k+1) = kappa(:,k)*ar(:)
    end do

    Cmt=0d0
    Smt=0d0

    call alfsect(nmax, gs, theta, pmm)

    do m=1,N1
      call alford(nmax, m-1, theta, pmm, root2, p)
      pk(:,:)=p(:,:)*kappa(:,:)
      Cmt(:,m)=matmul(pk(:,:),cnm(:,m))
      Smt(:,m)=matmul(pk(:,:),snm(:,m))
    end do

    do i = 1,KK1
      y(i) = 0d0
      do j = 1,N1
        lm = lambda(i) * dble(j-1)
        y(i) = y(i) + Cmt(i,j)*dcos(lm) + Smt(i,j)*dsin(lm)
      end do
      y(i) = y(i) / gs
    end do

    deallocate(root2)
    
  end function solid_polynomial_values_scattered


  !ALFSECT computes scaled point values of sectorial Associated Legendre Functions
  !
  ! INPUT
  ! 1. N is the maximum order of the required ALFs
  ! 2. gs is a global scaling factor for avoiding underflow in evaluating the
  !   Associated Legendre Functions, say gs=10^260.
  ! 3. theta is a vector with the co-latitudes
  ! 
  ! OUTPUT
  ! 1. pmm  is a matrix of size N+1 x (the lenth of  theta)  
  ! with the scaled values of sectorial ALF
  !
  ! Notes. 
  ! 1. ALFSECT follows Fortran alfpsec.f code
  ! 2. The input degree 'N' must be less than or equal to 2700.
  !
  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  subroutine alfsect(N,gs,theta, pmm)
    integer, intent(in) :: N
    double precision, intent(in) :: gs
    double precision, dimension(:), intent(in) :: theta
    double precision, dimension(N+1, size(theta)), intent(out) :: pmm

    integer :: K, n1, small, i, nu
    integer, dimension(size(theta)) :: zero
    double precision :: gslog, rn2, eps, val
    double precision, dimension(size(theta)) :: u
    double precision, dimension(N+1) :: c

    eps = 2.2204d-16

    K=size(theta)
    zero(:) = 0

    c(:) = 0d0
    c(2) = dsqrt(3d0)
    do n1 = 3,N+1
      rn2 = dble(n1-1)*2d0
      c(n1) = dsqrt(rn2+1d0)/dsqrt(rn2)
    end do

    small = -200
    gslog = dlog10(gs)

    u(:) = dsin(theta(:))

    do i = 1,K
      if(u(i)<eps) then
        zero(i)=1
      else if (u(i)>1d0-eps) then
        zero(i)=N+1
      else
        val=anint((dble(small) - gslog)/dlog10(u(i))) + 1d0
        if(val .gt. dble(N+1)) val = dble(N+1)
        zero(i) = int(val)
      end if
    end do

    pmm(:,:)=0d0

    do i=1,K
      pmm(1,i)=gs
      do nu=2,zero(i)
        pmm(nu,i)=c(nu)*u(i)*pmm(nu-1,i)
      end do
    end do

  end subroutine alfsect

  !ALFORD computes scaled point values of non-sectorial Associated Legendre
  ! Functions (ALF) for a given order 'm' and for all degrees from 'm' to 'N'
  !
  ! INPUT
  ! 1. N  is the maximum degree of the required ALFs
  ! 2. m  is the order for which non-sectorial ALFs are to be computed
  ! 3. theta is a vector with the co-latitudes
  ! 4. pmm  is a matrix of size N+1 x (the lenth of  theta)  
  !   with the scaled values of sectorial ALF
  ! 5. root2 is a work vector, root2=realsqrt(1:2*N+5)'
  !
  ! OUTPUT
  ! 4. p  is a matrix of size (the lenth of  theta) x N+1
  !   with the scaled values of non-sectorial ALF
  !
  ! Notes. 
  ! 1. ALFORD follows Fortran alfpord.f code
  ! 2. The input degree 'N' must be less than or equal to 2700.
  ! 3. The input order 'm' must be less than or equal to the maximum
  !   degree 'N'. 
  ! 4. Note that any global scaling of the computed ALFs enters through the
  !   pre-computed sectorial values in 'pmm'.
  !
  ! Version: May 2013
  ! Author:  Kamen Ivanov,      Contact: kamen@math.bas.bg
  !
  ! FORTRAN conversion: L. Scott Johnson <sjohnson.imi@gmail.com>

  subroutine alford(N,m,theta,pmm,root2, p)
    integer, intent(in) :: N, m
    double precision, dimension(:), intent(in) :: theta
    double precision, dimension(:,:), intent(in) :: pmm
    double precision, dimension(:), intent(in) :: root2

    double precision, dimension(size(theta), N+1) :: p

    integer :: K, n1
    double precision :: temp, a, b
    double precision, dimension(size(theta)) :: tc

    if (N>2700) then
      print *, 'ALFORD:big_degree'
      print *, 'Degree ', N, ' bigger than 2700'
      stop
    end if
    if (m>N) then
      print *, 'ALFORD:big_order'
      print *, 'Order ', m, ' bigger than degree ', N
      stop
    end if

    K=size(theta)
    tc(:) = dcos(theta(:))
    p(:,:) = 0d0

    p(:,m+1)=pmm(m+1,:)

    if (m .lt. N) p(:,m+2) = root2(2*m+3)*tc*p(:,m+1)

    do n1=m+3,N+1
      temp=root2(2*n1-1)/root2(n1-1+m)/root2(n1-1-m)
      a=root2(2*n1-3)*temp
      b=root2(n1+m-2)*root2(n1-m-2)*temp/root2(2*n1-5)
      p(:,n1)=a*tc*p(:,n1-1) - b*p(:,n1-2)
    end do

  end subroutine alford

  !      subroutine gd2gc(phi,ht,ae,b,phig,re)
  !     CONVERTS COORDINATES FROM GEODETIC (phi,ht) TO GEOCENTRIC (phig,re)
  !
  ! Remark. phi and phig are geograpic coordinates, i.e. angular distances to
  ! the equator (not to the notrh pole)!!!

  !       ae = SEMI-MAJOR AXIS .....................(m)
  !       b  = SEMI-MINOR AXIS .....................(m)
  !function [phig,re]=gd2gc(phi,ht,ae,b)
  subroutine gd2gc(phi,ht,ae,b,phig,re)
    double precision, dimension(:), intent(in) :: phi, ht
    double precision, intent(in) :: ae, b
    double precision, dimension(:), intent(out) :: phig, re

    integer :: sz
    double precision :: e2
    double precision, dimension(:), allocatable :: u, t, en, rho, z

    if (size(phi).ne.size(ht)) then
      print *,'GC2GE:phi_ht'
      print *,'phi and ht must be of the same size'
      stop
    end if
    if (size(phi).ne.size(phig)) then
      print *,'GC2GE:phi_phig'
      print *,'phi and phig must be of the same size'
      stop
    end if
    if (size(phi).ne.size(re)) then
      print *,'GC2GE:phi_re'
      print *,'phi and re must be of the same size'
      stop
    end if

    sz = size(phi)

    allocate(u(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: u'
      stop
    end if
    allocate(t(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: t'
      stop
    end if
    allocate(en(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: en'
      stop
    end if
    allocate(rho(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: rho'
      stop
    end if
    allocate(z(sz), stat=ret_i)
    if(ret_i .gt. 0) then
      print *, 'Out of memory: z'
      stop
    end if

    e2 = (ae**2-b**2)/(ae**2)
    u = dsin(phi)
    t = dcos(phi)
    en = ae/dsqrt(1d0 - e2*u**2)
    rho = (en + ht)*t
    z = (en*(1d0-e2) + ht)*u
    phig = datan2(z,rho)
    re = dsqrt(rho**2 + z**2)

    deallocate(z)
    deallocate(rho)
    deallocate(en)
    deallocate(t)
    deallocate(u)

  end subroutine gd2gc

  !----------------------------------------------------------------------------
  ! dlmsize: Count the lines in the dlm file
  integer function dlmsize(fname)

    character*(*), intent(in) :: fname
    integer n, ret_i

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read points file.'
      stop
    end if

    n = 0
    do
      read(1, *, iostat=ret_i) 
      if(ret_i .ne. 0) exit
      n = n + 1
    end do
    close(1)

    dlmsize = n
  end function dlmsize

  !----------------------------------------------------------------------------
  !  dlmread:   Reads scattered point coordinates from free-format ASCII file
  !     replaces the matlab function DLMREAD
  !     FORTRAN VERSION:                          L. Scott Johnson,  APR 2013
  subroutine dlmread(fname, n, ptarr)

    character*(*), intent(in) :: fname
    integer, intent(in) :: n
    double precision :: ptarr(n,3)
    integer :: i, ret_i

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read points file ', fname
      stop
    end if

    !--------------------------------------------------------------------------
    !     read the coordinates from the file into the array
    !--------------------------------------------------------------------------
    do i=1,n
      read(1,*) ptarr(i,1), ptarr(i,2), ptarr(i,3)
    end do
    close(1)
  end subroutine dlmread
  !----------------------------------------------------------------------------
  !  dlmread:   Reads scattered point coordinates from free-format ASCII file
  !     replaces the matlab function DLMREAD
  !     FORTRAN VERSION:                          L. Scott Johnson,  APR 2013
  subroutine dlmread13(fname, n, ptarr)

    character*(*), intent(in) :: fname
    integer, intent(in) :: n
    double precision :: ptarr(n,13)
    integer :: i, ret_i

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read points file ', fname
      stop
    end if

    !--------------------------------------------------------------------------
    !     read the coordinates from the file into the array
    !--------------------------------------------------------------------------
    do i=1,n
      read(1,*) ptarr(i,1:13)
    end do
    close(1)
  end subroutine dlmread13

  !----------------------------------------------------------------------------
  subroutine coeff_read(fname, cnm, snm)
    character*(*), intent(in) :: fname
    double precision, intent(out) :: cnm(:,:), snm(:,:)

    integer :: a, b, ret_i
    double precision :: c, s

    open(unit=1, file=fname, action='read', status='old',iostat=ret_i)
    if(ret_i .ne. 0) then
      print *, 'Cannot read coeff file ', fname
      stop
    end if

    do
      read(1,*,iostat=ret_i) a,b,c,s
      if(ret_i .ne. 0) exit
      cnm(a+1,b+1) = c
      snm(a+1,b+1) = s
    end do

    close(1)
  end subroutine coeff_read

end program emmsynth_standard
